# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class Distribution(ABC):
    """
    Base class for all distributions. Enforces that any created distributions
    have the required value method, which returns a singular value on request.
    If the distribution requires randomness, then the set_random_seed method
    should be added, which accepts a seed and sets this for the rng to allow
    for repeatability where required.
    """

    @abstractmethod
    def value(self) -> float:
        """Returns a value from the distribution on request."""

    @abstractmethod
    def set_random_seed(self, seed: int | None) -> None:
        """
        Used for setting the random seed for the model. Can just pass in cases
        where a distribution does not feature a random component.
        """

    def __repr__(self) -> str:
        return "lightworks.interferometers.dists." + str(self)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numpy import random

from .distribution import Distribution
from .utils import is_number


class TopHat(Distribution):
    """
    Returns random value according to a uniform distribution between two values.

    Args:

        min_value (float) : The minimum value of the distribution.

        max_value (float) : The maximum value of the distribution.

    """

    def __init__(self, min_value: float, max_value: float) -> None:
        # Type check all values
        is_number([min_value, max_value])
        # Then check min and max values are valid
        if max_value < min_value:
            raise ValueError("Max value cannot be less than min value.")
        # Then assign to attributes
        self._min_value = min_value
        self._max_value = max_value
        self._rng = random.default_rng()

    def __str__(self) -> str:
        return f"TopHat({self._min_value}, {self._max_value})"

    def value(self) -> float:
        """Returns random value from within set range."""
        return (
            self._min_value
            + (self._max_value - self._min_value) * self._rng.random()
        )

    def set_random_seed(self, seed: int | None) -> None:
        """Used for setting the random seed for the model."""
        self._rng = random.default_rng(seed)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

import drawsvg as draw
from multimethod import multimethod

from lightworks.sdk.circuit.photonic_components import (
    Barrier,
    BeamSplitter,
    Group,
    HeraldData,
    Loss,
    ModeSwaps,
    PhaseShifter,
    UnitaryMatrix,
)
from lightworks.sdk.utils.exceptions import DisplayError

from .display_utils import SVGSettings, process_parameter_value
from .draw_specs import (
    BarrierDrawing,
    BeamSplitterDrawing,
    DrawSpec,
    HeraldDrawing,
    LossDrawing,
    ModeSwapDrawing,
    PhaseShifterDrawing,
    TextDrawing,
    UnitaryDrawing,
    WaveguideDrawing,
)

if TYPE_CHECKING:
    from lightworks.sdk.circuit import PhotonicCircuit


class DrawCircuitSVG:
    """
    DrawCircuit

    This class can be used to Display a circuit in the quantum emulator as a
    figure in matplotlib.

    Args:

        circuit (PhotonicCircuit) : The circuit which is to be displayed.

        display_loss (bool, optional) : Choose whether to display loss
            components in the figure, defaults to False.

        mode_label (list|None, optional) : Optionally provided a list of mode
            labels which will be used to name the mode something other than
            numerical values. Can be set to None to use default values.

        show_parameter_values (bool, optional) : Shows the values of parameters
            instead of the associated labels if specified.

        display_barriers (bool, optional) : Shows included barriers within the
            created visualization if this is required.

    """

    def __init__(
        self,
        circuit: "PhotonicCircuit",
        display_loss: bool = False,
        mode_labels: list[str] | None = None,
        show_parameter_values: bool = False,
        display_barriers: bool = False,
    ) -> None:
        self.circuit = circuit
        self.display_loss = display_loss
        self.herald_modes = self.circuit._internal_modes
        self.show_parameter_values = show_parameter_values
        self.display_barriers = display_barriers
        # Set a waveguide width and get mode number
        self.wg_width = 8
        self.n_modes = self.circuit.n_modes
        self.draw_spec: list[DrawSpec] = []
        self.dy = 125
        self.dy_smaller = 75
        self.y_locations = []
        # Set mode y locations
        yloc = self.dy
        for i in range(self.n_modes):
            self.y_locations.append(yloc)
            if i + 1 in self.herald_modes or i in self.herald_modes:
                yloc += self.dy_smaller
            else:
                yloc += self.dy
        init_length = 100.0
        # Add the mode labels
        if mode_labels is not None:
            exp_len = self.n_modes - len(self.herald_modes)
            if len(mode_labels) != exp_len:
                msg = (
                    "Length of provided mode labels list should be equal to "
                    f"the number of useable modes ({exp_len})."
                )
                raise DisplayError(msg)
        else:
            mode_labels = [
                str(i) for i in range(self.n_modes - len(self.herald_modes))
            ]
        # Convert all labels to strings
        mode_labels = [str(m) for m in mode_labels]
        full_mode_labels = []
        count = 0
        for i in range(self.n_modes):
            if i not in self.herald_modes:
                full_mode_labels.append(mode_labels[count])
                count += 1
            else:
                full_mode_labels.append("-")
        mode_labels = full_mode_labels
        # Adjust canvas size for long labels
        max_len = max(len(m) for m in mode_labels)
        if max_len > 4:
            init_length += (max_len - 4) * 17.5
        for m, label in enumerate(mode_labels):
            self.draw_spec.append(
                TextDrawing(
                    text=label,
                    x=init_length - 20,
                    y=self.y_locations[m] + 2,
                    rotation=0,
                    size=SVGSettings.TEXT_SIZE.value,
                    colour="black",
                    alignment="right",
                )
            )

        self._init_length = init_length
        # Create list of locations for each mode
        self.x_locations = [init_length + 50] * self.n_modes
        # Add extra waveguides when using heralds
        if self.circuit._external_heralds.input:
            for m in range(self.n_modes):
                if m not in self.herald_modes:
                    self._add_wg(self.x_locations[m], self.y_locations[m], 50)
                    self.x_locations[m] += 50
        # Loop over each element in the circuit spec and add
        for spec in self.circuit._get_circuit_spec():
            self._add(spec)

        maxloc = max(self.x_locations)
        # Extend final waveguide if herald included
        if self.circuit._external_heralds.output:
            maxloc += 50
        for i, loc in enumerate(self.x_locations):
            if loc < maxloc and i not in self.herald_modes:
                self._add_wg(loc, self.y_locations[i], maxloc - loc)
                self.x_locations[i] = maxloc

        # Add heralding display
        self._add_heralds(
            self.circuit._external_heralds, self._init_length + 50, maxloc
        )

        for i in range(self.n_modes):
            self.x_locations[i] += 50

    def draw(self) -> draw.Drawing:
        """
        Draws the circuit as defined in the initial class class:

        Returns:

            draw.Drawing : The created drawing with all circuit components.

        """
        extra_lower = 0
        # Create a new drawing of the correct size
        self.d = draw.Drawing(
            max(self.x_locations) + 100,
            max(self.y_locations) + self.dy + extra_lower,
        )

        # Set white background
        self.d.append(
            draw.Rectangle(
                0, 0, self.d.width, self.d.height, fill="white", stroke="none"
            )
        )

        border = 100
        # Add frame around circuit
        dx = max(self.x_locations) - self._init_length
        dy = max(self.y_locations) - self.dy + 2 * border + extra_lower
        self.d.append(
            draw.Rectangle(
                self._init_length,
                self.dy - border,
                dx,
                dy,
                fill="none",
                stroke="black",
            )
        )
        # Add ticks to left side of frame
        length, width = 8, 1
        for i in range(self.n_modes):
            self.d.append(
                draw.Rectangle(
                    self._init_length - length,
                    self.y_locations[i] - width / 2,
                    length,
                    width,
                    fill="black",
                )
            )

        for spec in self.draw_spec:
            self.d.append(spec.draw_svg())

        mode_width = self.circuit.n_modes - 0.6 * len(
            self.circuit._internal_modes
        )
        # Adjust size of figure to meet target scale
        target_scale = min(50 + mode_width * 65, 900)
        self.d.set_pixel_scale(1)
        _w, h = self.d.calc_render_size()
        self.d.set_pixel_scale(target_scale / h)

        return self.d

    def _add_wg(self, x: float, y: float, length: float) -> None:
        """
        Add a waveguide to the drawing.
        """
        # Add a waveguide of the required length to the drawing spec
        self.draw_spec.append(
            WaveguideDrawing(
                x=x - 0.05, y=y, length=length + 0.1, width=self.wg_width
            )
        )

    @multimethod
    def _add(self, spec: Any) -> None:  # noqa: ARG002
        """
        Catch all for any components which may not have been implemented.
        """
        raise ValueError("Unrecognised component in circuit spec.")

    @_add.register
    def _add_ps(self, spec: PhaseShifter) -> None:
        """
        Add a phase shifter to the drawing.
        """
        phi = process_parameter_value(spec.phi, self.show_parameter_values)
        size = 50
        con_length = 50
        # Get current x and y locations for the mode
        xloc = self.x_locations[spec.mode]
        yloc = self.y_locations[spec.mode]
        # Input waveguide
        self._add_wg(xloc, yloc, con_length)
        xloc += con_length
        # Add phase shifter section
        self.draw_spec.append(
            PhaseShifterDrawing(x=xloc, y=yloc, size=size, phase=phi)
        )
        xloc += size
        # Output waveguide
        self._add_wg(xloc, yloc, con_length)
        self.x_locations[spec.mode] = xloc + con_length

    @_add.register
    def _add_bs(self, spec: BeamSplitter) -> None:
        """
        Add a beam splitter across to provided modes to the drawing.
        """
        mode1, mode2 = spec.mode_1, spec.mode_2
        ref = process_parameter_value(
            spec.reflectivity, self.show_parameter_values
        )
        if mode1 > mode2:
            mode1, mode2 = mode2, mode1
        size_x = 50  # x beam splitter size
        con_length = 50  # input/output waveguide length
        offset = 50  # Offset of beam splitter shape from mode centres
        if mode1 > mode2:
            mode1, mode2 = mode2, mode1
        size_y = offset + self.y_locations[mode2] - self.y_locations[mode1]
        # Get x and y locations
        yloc = self.y_locations[mode1]
        # Equalise all x waveguides
        xloc = self._equalize_waveguides(range(mode1, mode2 + 1))
        # Add input waveguides for all included modes
        for i in range(mode1, mode2 + 1):
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        # Check for any required text offset to avoid waveguides
        mode_between = 0
        for i in range(mode1 + 1, mode2 + 1, 1):
            if i not in self.herald_modes:
                mode_between += 1
                dy = -50 if mode_between % 2 == 0 else 0
        # Add beam splitter section
        self.draw_spec.append(
            BeamSplitterDrawing(
                x=xloc,
                y=yloc,
                size_x=size_x,
                size_y=size_y,
                offset_y=offset / 2,
                reflectivity=ref,
                text_offset=dy,
            )
        )
        # For any modes in between the beam splitter modes add a waveguide
        # across the beam splitter
        for i in range(mode1 + 1, mode2):
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], size_x)
        xloc += size_x
        # Add output waveguides and update mode locations
        for i in range(mode1, mode2 + 1):
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

    @_add.register
    def _add_unitary(self, spec: UnitaryMatrix) -> None:
        """
        Add a unitary matrix representation to the drawing.
        """
        mode1, mode2 = spec.mode, spec.mode + spec.unitary.shape[0] - 1
        size_x = 100  # Unitary x size
        con_length = 50  # Input/output waveguide lengths
        offset = 50  # Offset of unitary square from modes
        size_y = offset + self.y_locations[mode2] - self.y_locations[mode1]
        # Get x and y positions
        yloc = self.y_locations[mode1]
        # Equalise all x waveguides
        xloc = self._equalize_waveguides(range(mode1, mode2 + 1))
        # Add input waveguides
        modes = range(mode1, mode2 + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        # Add unitary shape and U label
        self.draw_spec.append(
            UnitaryDrawing(
                x=xloc,
                y=yloc,
                size_x=size_x,
                size_y=size_y,
                offset_y=offset / 2,
                label=spec.label,
                text_size=SVGSettings.TEXT_SIZE.value
                if self.n_modes > 2
                else SVGSettings.U_TEXT_SIZE.value,
            )
        )
        xloc += size_x
        # Add output waveguides and update mode positions
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

    @_add.register
    def _add_loss(self, spec: Loss) -> None:
        """
        Add a loss channel to the specified mode.
        """
        # Don't add if not enabled
        if not self.display_loss:
            return
        loss = process_parameter_value(spec.loss, self.show_parameter_values)
        size = 50
        con_length = 50
        # Get current x and y locations for the mode
        xloc = self.x_locations[spec.mode]
        yloc = self.y_locations[spec.mode]
        # Input waveguide
        self._add_wg(xloc, yloc, con_length)
        xloc += con_length
        # Add phase shifter section
        self.draw_spec.append(LossDrawing(x=xloc, y=yloc, size=size, loss=loss))
        xloc += size
        # Output waveguide
        self._add_wg(xloc, yloc, con_length)
        self.x_locations[spec.mode] = xloc + con_length

        return

    @_add.register
    def _add_barrier(self, spec: Barrier) -> None:
        """
        Add a barrier which will separate different parts of the circuit. This
        is applied to the provided modes.
        """
        max_loc = max(self.x_locations[m] for m in spec.modes)
        for m in spec.modes:
            loc = self.x_locations[m]
            if loc < max_loc:
                self._add_wg(loc, self.y_locations[m], max_loc - loc)
            self.x_locations[m] = max_loc
            if self.display_barriers:
                self.draw_spec.append(
                    BarrierDrawing(
                        x=max_loc,
                        y_start=self.y_locations[m] - self.dy_smaller / 2,
                        y_end=self.y_locations[m] + self.dy_smaller / 2,
                        width=self.wg_width * 0.6,
                    )
                )

    @_add.register
    def _add_mode_swaps(self, spec: ModeSwaps) -> None:
        """
        Add mode swaps between provided modes to the drawing.
        """
        # Skip any empty elements
        if not spec.swaps:
            return
        swaps = dict(spec.swaps)  # Make copy of swaps
        con_length = 25  # input/output waveguide length
        min_mode = min(swaps)
        max_mode = max(swaps)
        size_x = 50 + 20 * (max_mode - min_mode)  # x length of swap element
        # Add in missing mode for swap
        for m in range(min_mode, max_mode + 1):
            if m not in swaps:
                swaps[m] = m
        # Get y locations
        ylocs = []
        for i, j in swaps.items():
            if i not in self.herald_modes:
                ylocs.append(
                    (float(self.y_locations[i]), float(self.y_locations[j]))
                )
        # Equalise all x waveguides
        xloc = self._equalize_waveguides(range(min_mode, max_mode + 1))
        # Add input waveguides for all included modes
        modes = range(min_mode, max_mode + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        # Add beam splitter section
        self.draw_spec.append(
            ModeSwapDrawing(
                x=xloc, ys=ylocs, size_x=size_x, wg_width=self.wg_width
            )
        )
        xloc += size_x
        # Add output waveguides update mode locations
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

        return

    @_add.register
    def _add_grouped_circuit(self, spec: Group) -> None:
        """
        Add a grouped_circuit representation to the drawing.
        """
        mode1, mode2 = spec.mode_1, spec.mode_2
        if mode1 > mode2:
            mode1, mode2 = mode2, mode1
        size_x = 100  # Drawing x size
        con_length = 50  # Input/output waveguide lengths
        extra_length = 50 if spec.heralds.input or spec.heralds.output else 0
        offset = 50  # Offset of square from modes
        size_y = offset + self.y_locations[mode2] - self.y_locations[mode1]
        # Get x and y positions
        yloc = self.y_locations[mode1]
        # Equalise all x waveguides
        xloc = self._equalize_waveguides(range(mode1, mode2 + 1))
        # Add input waveguides
        modes = range(mode1, mode2 + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(
                    xloc, self.y_locations[i], con_length + extra_length
                )
            elif i - mode1 in spec.heralds.input:
                self._add_wg(
                    xloc + extra_length, self.y_locations[i], con_length
                )
        xloc += con_length + extra_length
        # Add unitary shape and U label
        self.draw_spec.append(
            UnitaryDrawing(
                x=xloc,
                y=yloc,
                size_x=size_x,
                size_y=size_y,
                offset_y=offset / 2,
                label=spec.name,
                text_size=SVGSettings.TEXT_SIZE.value
                if self.n_modes > 2
                else SVGSettings.U_TEXT_SIZE.value,
            )
        )
        xloc += size_x
        # Add output waveguides and update mode positions
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(
                    xloc, self.y_locations[i], con_length + extra_length
                )
            elif i - mode1 in spec.heralds.output:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length + extra_length

        # Modify provided heralds by mode offset and then add at locations
        shifted_heralds = HeraldData(
            input={m + mode1: n for m, n in spec.heralds.input.items()},
            output={m + mode1: n for m, n in spec.heralds.output.items()},
        )
        self._add_heralds(
            shifted_heralds, xloc - size_x - con_length, xloc + con_length
        )

    def _add_heralds(
        self,
        heralds: HeraldData,
        start_loc: float,
        end_loc: float,
    ) -> None:
        """
        Adds display of all heralds to circuit.
        """
        size = 25
        # Input heralds
        for mode, num in heralds.input.items():
            xloc = start_loc
            yloc = self.y_locations[mode]
            self.draw_spec.append(
                HeraldDrawing(x=xloc, y=yloc, size=size, n_photons=num)
            )
        # Output heralds
        for mode, num in heralds.output.items():
            xloc = end_loc
            yloc = self.y_locations[mode]
            self.draw_spec.append(
                HeraldDrawing(x=xloc, y=yloc, size=size, n_photons=num)
            )

    def _equalize_waveguides(self, modes: Iterable[int]) -> float:
        """
        Extends the waveguide of any modes in the list to the target position.
        """
        all_locations = [self.x_locations[m] for m in modes]
        xloc = max(all_locations)
        for mode, loc in zip(modes, all_locations, strict=True):
            if loc < xloc and mode not in self.herald_modes:
                self._add_wg(loc, self.y_locations[mode], xloc - loc)
        return xloc



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from lightworks import ModeRangeError, Parameter, random_unitary
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.circuit.photonic_components import (
    Barrier,
    BeamSplitter,
    Group,
    Loss,
    ModeSwaps,
    PhaseShifter,
    UnitaryMatrix,
)


class TestCompiledCircuit:
    """
    Unit tests to confirm correct functioning of the CompiledPhotonicCircuit
    class when various operations are performed.
    """

    def test_creation(self):
        """
        Checks creation of a compiled circuit doesn't raise any issues.
        """
        CompiledPhotonicCircuit(10)

    def test_creation_modes(self):
        """
        Checks number of modes are correct on creation of a circuit.
        """
        c = CompiledPhotonicCircuit(10)
        assert c.n_modes == 10
        assert c.loss_modes == 00
        assert c.n_modes == c.total_modes

    def test_default_array_type(self):
        """
        Checks that default array type is complex.
        """
        c = CompiledPhotonicCircuit(10)
        assert c.U_full.dtype == complex

    @pytest.mark.parametrize(
        "component",
        [
            Barrier(list(range(4))),
            BeamSplitter(1, 3, 0.4, "Rx"),
            Loss(1, 0.5),
            ModeSwaps({0: 2, 2: 1, 1: 0}),
            PhaseShifter(3, 0.6),
            UnitaryMatrix(1, random_unitary(4), ""),
        ],
    )
    def test_component_addition(self, component):
        """
        Confirms each component can be added to a circuit and the configured
        unitary matches that from the component.
        """
        c = CompiledPhotonicCircuit(5)
        c.add(component)
        assert (
            c.U_full.round(8) == component.get_unitary(c.total_modes).round(8)
        ).all()

    def test_component_addition_group(self):
        """
        Confirms a group can be added to the circuit and that the unitary
        matches that which is manually calculated.
        """
        spec = [
            BeamSplitter(1, 2, 0.6, "Rx"),
            PhaseShifter(3, 0.6),
            Loss(1, 0.5),
            ModeSwaps({0: 2, 2: 1, 1: 0}),
        ]
        group = Group(spec, "test", 1, 3, {})
        c = CompiledPhotonicCircuit(5)
        c.add(group)
        # Find expected unitary from components individually
        expected = np.identity(c.total_modes)
        for s in spec:
            expected = s.get_unitary(c.total_modes) @ expected
        assert (c.U_full.round(8) == expected.round(8)).all()

    def test_parameter_support(self):
        """
        Confirms a parameter can be assigned to a component and it changes the
        produced circuit.
        """
        param = Parameter(0.9)
        bs = BeamSplitter(1, 3, param, "H")
        # Get first unitary
        c = CompiledPhotonicCircuit(4)
        c.add(bs)
        u1 = c.U_full
        # Then update parameter and get second
        param.set(0.3)
        c = CompiledPhotonicCircuit(4)
        c.add(bs)
        u2 = c.U_full
        assert (u1.round(8) != u2.round(8)).any()

    def test_herald(self):
        """
        Confirms that heralding being added to a circuit works as expected and
        is reflected in the heralds attribute.
        """
        circuit = CompiledPhotonicCircuit(4)
        circuit.add_herald(0, 2, 1, 3)
        # Check heralds added
        assert 0 in circuit.heralds.input
        assert 2 in circuit.heralds.output
        # Check photon number is correct
        assert circuit.heralds.input[0] == 1
        assert circuit.heralds.output[2] == 3

    @pytest.mark.parametrize("value", [2.5, "2", True])
    def test_herald_invalid_photon_number(self, value):
        """
        Checks error is raised when a non-integer photon number is provided to
        the herald method.
        """
        circuit = CompiledPhotonicCircuit(4)
        with pytest.raises(TypeError):
            circuit.add_herald(0, 0, value, 0)

    @pytest.mark.parametrize("value", [2.5, 6, True])
    def test_herald_invalid_mode_number(self, value):
        """
        Checks error is raised when a non-integer mode number is provided to
        the herald method.
        """
        circuit = CompiledPhotonicCircuit(4)
        with pytest.raises((TypeError, ModeRangeError)):
            circuit.add_herald(value, 0, 0, 0)

    @pytest.mark.parametrize("value", [2.5, 6, True])
    def test_herald_invalid_mode_number_output(self, value):
        """
        Checks error is raised when a non-integer output mode number is provided
        to the herald method.
        """
        circuit = CompiledPhotonicCircuit(4)
        with pytest.raises((TypeError, ModeRangeError)):
            circuit.add_herald(0, value, 0, 0)

    def test_herald_duplicate_value(self):
        """
        Checks error is raised if a duplicate mode is provided to the herald.
        """
        circuit = CompiledPhotonicCircuit(4)
        circuit.add_herald(0, 0, 0, 0)
        with pytest.raises(ValueError):
            circuit.add_herald(0, 0, 0, 0)

    def test_herald_duplicate_value_output(self):
        """
        Checks error is raised if a duplicate output mode is provided to the
        herald.
        """
        circuit = CompiledPhotonicCircuit(4)
        circuit.add_herald(0, 1, 0, 0)
        with pytest.raises(ValueError):
            circuit.add_herald(1, 1, 0, 0)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

import numpy as np
from numpy.typing import NDArray

# NOTE: This file is auto-documented, so any functions not intended to be public
# should begin with _

T = TypeVar("T", bound=np.generic)


def project_density_to_physical(
    rho: NDArray[np.complex128],
) -> NDArray[np.complex128]:
    """
    Takes a provided density matrix and projects into to a physical space using
    the algorithm from https://doi.org/10.48550/arXiv.1106.5458. It ensures:

        1. The matrix is Hermitian
        2. The trace of the matrix is 1.
        3. All eigenvalues of the matrix are positive, if not the matrix is
           projected so this is the case.

    Args:

        rho (np.ndarray) : The matrix which is to be projected.

    Returns:

        np.ndarray : The calculated density matrix, which meets the required
            conditions to be physical.

    """
    dim = rho.shape[0]
    # Ensure matrix is hermitian
    rho = (rho + rho.conj().T) / 2
    # And that the trace is 1
    rho /= np.trace(rho)
    # Then find eigenvalues and eigenvectors
    eigvals, eigvecs = _find_eigen_data(rho)
    # Don't do anything else if the eigenvalues are all positive
    if min(eigvals) >= 0:
        return rho
    # Otherwise apply algorithm - requires eigenvalues are sorted from largest
    # to smallest so invert these first
    eigvals = np.flip(eigvals)
    i = dim
    acc = 0.0
    while eigvals[i - 1] + acc / i < 0:
        acc += eigvals[i - 1]
        i -= 1
    for j in range(dim):
        if j < i:
            eigvals[j] += acc / i
        else:
            eigvals[j] = 0
    # Return eigenvalues to the correct order
    eigvals = np.flip(eigvals)
    # Reconstruct the matrix and return
    return eigvecs @ np.diag(eigvals) @ np.conj(eigvecs.T)


def project_choi_to_physical(
    choi: NDArray[np.complex128], max_iter: int = 1000
) -> NDArray[np.complex128]:
    """
    Performs the CPTP algorithm from https://arxiv.org/abs/1803.10062, ensuring
    the provided choi matrix is completely positive and trace preserving. This
    uses a series of repeat projection steps until a convergence is reached.

    Args:

        choi (np.ndarray) : The matrix which is to be projected.

        max_iter (int) : The max number of iterations that can be performed by
            the projection algorithm.

    Returns:

        np.ndarray : The calculated choi matrix, which meets the required
            conditions to be physical.

    """
    x_0 = _vec(choi)
    dim = x_0.shape[0]
    # Initialize most quantities as a zero matrix
    p_0 = np.array([0] * dim)
    q_0 = np.array([0] * dim)
    y_0 = np.array([0] * dim)
    # Run for maximum number of iterations
    for _ in range(max_iter):
        # Calculate updated quantiles
        y_k = _vec(_tp_proj(_unvec(x_0 + p_0)))
        p_k = x_0 + p_0 - y_k
        x_k = _vec(_cp_proj(_unvec(y_k + q_0)))
        q_k = y_k + q_0 - x_k
        # Stopping condition (see paper)
        if (
            np.linalg.norm(p_0 - p_k) ** 2
            + np.linalg.norm(q_0 - q_k) ** 2
            + abs(2 * np.conj(p_0) @ (x_k - x_0))
            + abs(2 * np.conj(q_0) @ (y_k - y_0))
        ) < 10**-4:
            break
        # Update values
        y_0, p_0, x_0, q_0 = y_k, p_k, x_k, q_k

    return _unvec(x_k)


def _cp_proj(choi: NDArray[np.complex128]) -> NDArray[np.complex128]:
    """
    Performs the CP part of the projection by enforcing that the eigenvalues
    of the matrix are all positive.
    """
    vals, vecs = _find_eigen_data(choi)
    d = np.diag([max(v, 0) for v in vals])
    return vecs @ d @ np.conj(vecs.T)


def _tp_proj(choi: NDArray[np.complex128]) -> NDArray[np.complex128]:
    """
    Performs the TP part of the projection by enforcing that the partial
    trace of the choi matrix is equal to the identity matrix.
    """
    # Compute partial trace
    dim = int(choi.shape[0] ** 0.5)
    partial_trace = choi.reshape(np.tile([dim, dim], 2))
    partial_trace = np.einsum(partial_trace, np.array([0, 1, 2, 1])).astype(  # type: ignore[arg-type]
        complex
    )
    partial_trace = partial_trace.reshape(dim, dim)  # type: ignore[assignment]
    # Then find the variation and subtract this from the choi matrix
    variation = partial_trace - np.identity(dim)
    return choi - np.kron(variation / dim, np.identity(dim))


def _find_eigen_data(
    matrix: NDArray[np.complex128],
) -> tuple[NDArray[np.float64], NDArray[np.complex128]]:
    """
    Calculates and returns the eigenvalues and eigenvectors of a provided
    matrix. These are returned from the smallest to the largest eigenvalue.
    """
    return np.linalg.eigh(matrix)


def _vec(mat: NDArray[T]) -> NDArray[T]:
    """
    Applies flatten operation to a provided matrix to convert it into a vector.
    """
    return mat.flatten()


def _unvec(mat: NDArray[T]) -> NDArray[T]:
    """
    Takes a provided vector and converts it into a square matrix.
    """
    dim = int(mat.shape[0] ** 0.5)
    return mat.reshape(dim, dim)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import (
    ModeMismatchError,
    Parameter,
    PhotonicCircuit,
    PhotonNumberError,
    PostSelection,
    Sampler,
    State,
    Unitary,
    convert,
    random_unitary,
)
from lightworks.emulator import Backend, Detector, Source

P_BACKEND = Backend("permanent")


class TestSamplerGeneral:
    """
    Unit tests to check non-backend specific functionality of Sampler.
    """

    def test_sample_n_states_seed(self):
        """
        Checks that two successive function calls with a consistent seed
        produce the same result.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 5000, random_seed=1)
        results = P_BACKEND.run(sampler)
        results2 = P_BACKEND.run(sampler)
        assert results == results2

    def test_sample_n_states_seed_detector(self):
        """
        Checks that two successive function calls with a consistent seed
        produce the same result when an imperfect detector is used.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(
            circuit,
            State([1, 0, 1, 0]),
            5000,
            detector=Detector(efficiency=0.5, p_dark=1e-3),
            random_seed=1,
            sampling_mode="input",
        )
        results = P_BACKEND.run(sampler)
        results2 = P_BACKEND.run(sampler)
        assert results == results2

    def test_circuit_update_with_sampler(self):
        """
        Checks that when a circuit is modified then the sampler recalculates
        the probability distribution.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        circuit.bs(0)
        circuit.bs(2)
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2

    def test_circuit_parameter_update_with_sampler(self):
        """
        Checks that when the parameters of a circuit are updated then the
        corresponding probability distribution is modified.
        """
        p = Parameter(0.3)
        circuit = PhotonicCircuit(4)
        circuit.bs(0, reflectivity=p)
        circuit.bs(2, reflectivity=p)
        circuit.bs(1, reflectivity=p)
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        p.set(0.7)
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2

    def test_input_update_with_sampler(self):
        """
        Confirms that changing the input state to the sampler alters the
        produced results.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        sampler.input_state = State([0, 1, 0, 1])
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2

    def test_circuit_assignment(self):
        """
        Confirms that a PhotonicCircuit cannot be replaced with a
        non-PhotonicCircuit through the circuit attribute.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        with pytest.raises(TypeError):
            sampler.circuit = random_unitary(4)

    def test_input_assignmnet(self):
        """
        Checks that the input state of the sampler cannot be assigned to a
        non-State value and requires the correct number of modes.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        # Incorrect type
        with pytest.raises(TypeError):
            sampler.input_state = [1, 2, 3, 4]
        # Incorrect number of modes
        with pytest.raises(ModeMismatchError):
            sampler.input_state = State([1, 2, 3])

    def test_source_assignment(self):
        """
        Confirms that a Source cannot be replaced with a non-source object
        through the source attribute.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        with pytest.raises(TypeError):
            sampler.source = random_unitary(4)

    def test_detector_assignment(self):
        """
        Confirms that a Detector cannot be replaced with a non-detector object
        through the detector attribute.
        """
        circuit = Unitary(random_unitary(4))
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        with pytest.raises(TypeError):
            sampler.detector = random_unitary(4)

    def test_imperfect_source_update_with_sampler(self):
        """
        Checks that updating the parameters of a source will alter the
        calculated probability distribution from the sampler.
        """
        circuit = Unitary(random_unitary(4))
        source = Source(
            brightness=0.9,
            purity=0.9,
            indistinguishability=0.9,
            probability_threshold=1e-6,
        )
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000, source=source)
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        # Indistinguishability
        source.indistinguishability = 0.2
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2
        # Purity (reset previous variable to original value)
        source.indistinguishability = 0.9
        source.purity = 0.7
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2
        # Brightness
        source.purity = 0.9
        source.brightness = 0.4
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2
        # Probability threshold
        source.brightness = 0.9
        source.probability_threshold = 1e-3
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 != p2
        # Return all values to defaults and check this then returns the
        # original distribution
        source.probability_threshold = 1e-6
        P_BACKEND.run(sampler)
        p2 = sampler.probability_distribution
        assert p1 == p2

    def test_slos_equivalence_basic(self):
        """
        Checks probability distribution calculation from a simple unitary is
        nearly equivalent using both permanent and slos calculations.
        """
        circuit = Unitary(random_unitary(4))
        # Permanent
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        # SLOS
        Backend("slos").run(sampler)
        p2 = sampler.probability_distribution
        for s in p1:
            if round(p1[s], 8) != round(p2[s], 8):
                pytest.fail("Methods do not produce equivalent distributions.")

    def test_slos_equivalence_complex(self):
        """
        Checks probability distribution calculation is nearly equivalent using
        both permanent and slos calculations, when using loss and an imperfect
        source.
        """
        circuit = Unitary(random_unitary(4))
        for i in range(4):
            circuit.loss(i, 1)
        source = Source(indistinguishability=0.9, brightness=0.9, purity=0.9)
        # Permanent
        sampler = Sampler(
            circuit,
            State([1, 0, 1, 0]),
            1000,
            source=source,
        )
        P_BACKEND.run(sampler)
        p1 = sampler.probability_distribution
        # SLOS
        Backend("slos").run(sampler)
        p2 = sampler.probability_distribution
        # Test equivalence
        for s in p1:
            if round(p1[s], 8) != round(p2[s], 8):
                pytest.fail("Methods do not produce equivalent distributions.")

    def test_backend_updated_recalculates(self):
        """
        Checks that updating the backend causes recalculation of the
        probability distribution. This is achieved by checking a cache doesn't
        exist on the new backend.
        """
        circuit = Unitary(random_unitary(4))
        # Get initial distribution
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        P_BACKEND.run(sampler)
        b2 = Backend("slos")
        # Check attribute doesn't exist
        assert not hasattr(b2._backend, "_cache")

    def test_output_heralds_too_large(self):
        """
        Confirms a PhotonNumberError is raised when the number of output heralds
        is larger than the number of photons input into the system.
        """
        circuit = Unitary(random_unitary(4))
        circuit.herald(3, (0, 2))
        sampler = Sampler(circuit, State([1, 0, 0]), 1000)
        with pytest.raises(PhotonNumberError):
            sampler._generate_task()


@pytest.mark.parametrize("backend", [Backend("permanent"), Backend("slos")])
class TestSamplerCalculationBackends:
    """
    Unit tests to check results produced by Sampler object in the emulator with
    both SLOS and permanent backends.
    """

    def test_hom(self, backend):
        """
        Checks sampling a basic 2 photon input onto a 50:50 beam splitter,
        which should undergo HOM, producing outputs of |2,0> and |0,2>.
        """
        circuit = PhotonicCircuit(2)
        circuit.bs(0)
        n_sample = 100000
        sampler = Sampler(circuit, State([1, 1]), n_sample, random_seed=21)
        results = backend.run(sampler)
        assert len(results) == 2
        assert 0.49 < results[State([2, 0])] / n_sample < 0.51
        assert 0.49 < results[State([0, 2])] / n_sample < 0.51

    def test_hom_sample_n_outputs(self, backend):
        """
        Checks a lossy hom experiment with sample N outputs produces outputs of
        |2,0> and |0,2>.
        """
        circuit = PhotonicCircuit(2)
        circuit.bs(0, loss=0.1)
        n_sample = 100000
        sampler = Sampler(
            circuit, State([1, 1]), n_sample, random_seed=54, min_detection=2
        )
        results = backend.run(sampler)
        assert sum(results.values()) == n_sample
        assert 0.49 < results[State([2, 0])] / n_sample < 0.51
        assert 0.49 < results[State([0, 2])] / n_sample < 0.51

    def test_known_result(self, backend):
        """
        Builds a circuit which produces a known result and checks this is found
        at the output.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(1)
        circuit.mode_swaps({0: 1, 1: 0, 2: 3, 3: 2})
        circuit.bs(0, 3)
        # And check output counts
        sampler = Sampler(circuit, State([1, 0, 0, 1]), 1000)
        results = backend.run(sampler)
        assert results[State([0, 1, 1, 0])] == 1000

    def test_known_result_single_sample(self, backend):
        """
        Builds a circuit which produces a known result and checks this is found
        at the output, when using the sample method to get a single value.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(1)
        circuit.mode_swaps({0: 1, 1: 0, 2: 3, 3: 2})
        circuit.bs(0, 3)
        # And check output counts
        sampler = Sampler(circuit, State([1, 0, 0, 1]), 1)
        output = next(iter(backend.run(sampler)))
        assert output == State([0, 1, 1, 0])

    def test_sampling_perfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler is
        correct for a perfect source.
        """
        unitary = Unitary(random_unitary(4, seed=20))
        sampler = Sampler(unitary, State([1, 0, 1, 0]), 1000)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 1, 1, 0])]
        assert p == pytest.approx(0.112093500, 1e-8)

    def test_sampling_imperfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler is
        correct for an imperfect source.
        """
        unitary = Unitary(random_unitary(4, seed=20))
        source = Source(purity=0.9, brightness=0.9, indistinguishability=0.9)
        sampler = Sampler(unitary, State([1, 0, 1, 0]), 1000, source=source)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 0, 1, 0])]
        assert p == pytest.approx(0.0129992654, 1e-8)

    def test_sampling_2photons_in_mode_perfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler is
        correct for a perfect source with 2 photons in single input mode.
        """
        unitary = Unitary(random_unitary(4, seed=20))
        sampler = Sampler(unitary, State([0, 2, 0, 0]), 1000)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 1, 1, 0])]
        assert p == pytest.approx(0.2875114938, 1e-8)

    def test_sampling_2photons_in_mode_imperfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler is
        correct for an imperfect source with 2 photons in single input mode.
        """
        unitary = Unitary(random_unitary(4, seed=20))
        source = Source(purity=0.9, brightness=0.9, indistinguishability=0.9)
        sampler = Sampler(unitary, State([0, 2, 0, 0]), 1000, source=source)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 0, 1, 0])]
        assert p == pytest.approx(0.09767722765, 1e-8)

    def test_lossy_sampling_perfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler with
        a lossy circuit is correct for a perfect source.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(0, loss=convert.db_loss_to_decimal(1))
        circuit.bs(2, loss=convert.db_loss_to_decimal(2))
        circuit.ps(1, 0.3, loss=convert.db_loss_to_decimal(0.5))
        circuit.ps(3, 0.3, loss=convert.db_loss_to_decimal(0.5))
        circuit.bs(1, loss=convert.db_loss_to_decimal(1))
        circuit.bs(2, loss=convert.db_loss_to_decimal(2))
        circuit.ps(1, 0.3, loss=convert.db_loss_to_decimal(0.5))
        # Sample from circuit
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 1, 1, 0])]
        assert p == pytest.approx(0.01111424631, 1e-8)
        p = sampler.probability_distribution[State([0, 0, 0, 0])]
        assert p == pytest.approx(0.24688532527, 1e-8)

    def test_lossy_sampling_imperfect_source(self, backend):
        """
        Checks that the probability distribution calculated by the sampler with
        a lossy circuit is correct for an imperfect source.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(0, loss=convert.db_loss_to_decimal(1))
        circuit.bs(2, loss=convert.db_loss_to_decimal(2))
        circuit.ps(1, 0.3, loss=convert.db_loss_to_decimal(0.5))
        circuit.ps(3, 0.3, loss=convert.db_loss_to_decimal(0.5))
        circuit.bs(1, loss=convert.db_loss_to_decimal(1))
        circuit.bs(2, loss=convert.db_loss_to_decimal(2))
        circuit.ps(1, 0.3, loss=convert.db_loss_to_decimal(0.5))
        # Sample from circuit
        source = Source(purity=0.9, brightness=0.9, indistinguishability=0.9)
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000, source=source)
        backend.run(sampler)
        p = sampler.probability_distribution[State([0, 0, 1, 0])]
        assert p == pytest.approx(0.03122592963, 1e-8)
        p = sampler.probability_distribution[State([0, 0, 0, 0])]
        assert p == pytest.approx(0.28709359025, 1e-8)

    def test_imperfect_detection(self, backend):
        """Tests the behaviour of detectors with less than ideal efficiency."""
        circuit = Unitary(random_unitary(4))
        # Control
        detector = Detector(efficiency=1)
        sampler = Sampler(
            circuit,
            State([1, 0, 1, 0]),
            1000,
            detector=detector,
            sampling_mode="input",
        )
        results = backend.run(sampler)
        undetected_photons = False
        for s in results:
            if s.n_photons < 2:
                undetected_photons = True
                break
        assert not undetected_photons
        # With lossy detector
        detector = Detector(efficiency=0.5)
        sampler.detector = detector
        results = backend.run(sampler)
        undetected_photons = False
        for s in results:
            if s.n_photons < 2:
                undetected_photons = True
                break
        assert undetected_photons

    def test_detector_dark_counts(self, backend):
        """Confirms detector dark counts are introduced as expected."""
        circuit = Unitary(random_unitary(4))
        # Control
        detector = Detector(p_dark=0)
        sampler = Sampler(
            circuit,
            State([0, 0, 0, 0]),
            1000,
            detector=detector,
            sampling_mode="input",
        )
        results = backend.run(sampler)
        dark_counts = False
        for s in results:
            if s.n_photons > 0:
                dark_counts = True
                break
        assert not dark_counts
        # With dark counts enabled
        detector = Detector(p_dark=0.1)
        sampler.detector = detector
        results = backend.run(sampler)
        dark_counts = False
        for s in results:
            if s.n_photons > 0:
                dark_counts = True
                break
        assert dark_counts

    def test_detector_photon_counting(self, backend):
        """
        Checks that detector photon counting control alters the output states
        as expected after sampling.
        """
        circuit = Unitary(random_unitary(4))
        # Photon number resolving
        detector = Detector(photon_counting=True)
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000, detector=detector)
        results = backend.run(sampler)
        all_2_photon_states = True
        for s in results:
            if s.n_photons < 2:
                all_2_photon_states = False
                break
        assert all_2_photon_states
        # Non-photon number resolving
        detector = Detector(photon_counting=False)
        sampler.detector = detector
        results = backend.run(sampler)
        sub_2_photon_states = False
        for s in results:
            if s.n_photons < 2:
                sub_2_photon_states = True
                break
        assert sub_2_photon_states

    @pytest.mark.parametrize("n_output", [0, 1])
    @pytest.mark.flaky(reruns=3)
    def test_herald_equivalent(self, backend, n_output):
        """
        Checks that results are equivalent if a herald is used vs
        post-selection on a non-heralded circuit.
        """
        circuit = Unitary(random_unitary(6))
        # Sampler without built-in heralds
        sampler = Sampler(
            circuit,
            State([1, 1, 0, 1, 0, 0]),
            50000,
            post_selection=lambda s: s[1] == 1 and s[3] == n_output,
        )
        results = backend.run(sampler)
        # Then add and re-sample
        circuit.herald((0, 1), 1)
        circuit.herald((2, 3), (0, n_output))
        sampler = Sampler(circuit, State([1, 1, 0, 0]), 50000)
        results2 = backend.run(sampler)
        # Check all results with outputs greater than 2000
        for s, c in results2.items():
            if c > 2000:
                full_s = (
                    s[0:1] + State([1]) + s[1:2] + State([n_output]) + s[2:]
                )
                # Check results are within 10%
                assert pytest.approx(results[full_s], 0.1) == c

    @pytest.mark.parametrize("n_output", [0, 1])
    @pytest.mark.flaky(reruns=3)
    def test_herald_equivalent_grouped(self, backend, n_output):
        """
        Checks that results are equivalent if a herald is used vs
        post-selection on a non-heralded circuit when a grouped circuit is used.
        """
        circuit = Unitary(random_unitary(6))
        # Sampler without built-in heralds
        sampler = Sampler(
            circuit,
            State([1, 1, 0, 1, 0, 0]),
            50000,
            post_selection=lambda s: s[1] == 1 and s[3] == n_output,
        )
        results = backend.run(sampler)
        # Then add and re-sample
        circuit.herald((0, 1), 1)
        circuit.herald((2, 3), (0, n_output))
        full_circ = PhotonicCircuit(4)
        full_circ.add(circuit)
        sampler = Sampler(full_circ, State([1, 1, 0, 0]), 50000)
        results2 = backend.run(sampler)
        # Check all results with outputs greater than 2000
        for s, c in results2.items():
            if c > 2000:
                full_s = (
                    s[0:1] + State([1]) + s[1:2] + State([n_output]) + s[2:]
                )
                # Check results are within 10%
                assert pytest.approx(results[full_s], 0.1) == c

    @pytest.mark.flaky(reruns=2)
    def test_herald_equivalent_imperfect_source(self, backend):
        """
        Checks that results are equivalent if a herald is used vs
        post-selection on a non-heralded circuit, while also including an
        imperfect photon source.
        """
        circuit = Unitary(random_unitary(6))
        # Define source to use
        source = Source(purity=0.9, brightness=0.9, indistinguishability=0.9)
        # Sampler without built-in heralds
        sampler = Sampler(
            circuit,
            State([1, 1, 0, 1, 0, 0]),
            50000,
            source=source,
            post_selection=lambda s: s[1] == 1 and s[3] == 0,
        )
        results = backend.run(sampler)
        # Then add and re-sample
        circuit.herald((0, 1), 1)
        circuit.herald((2, 3), 0)
        sampler = Sampler(circuit, State([1, 1, 0, 0]), 50000, source=source)
        results2 = backend.run(sampler)
        # Check all results with outputs greater than 2000
        for s, c in results2.items():
            if c > 2000:
                full_s = s[0:1] + State([1]) + s[1:2] + State([0]) + s[2:]
                # Check results are within 10%
                assert pytest.approx(results[full_s], 0.1) == c

    @pytest.mark.flaky(reruns=2)
    def test_herald_equivalent_lossy(self, backend):
        """
        Checks that results are equivalent if a herald is used vs
        post-selection on a non-heralded lossy circuit.
        """
        circuit = Unitary(random_unitary(6))
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        # Sampler without built-in heralds
        sampler = Sampler(
            circuit,
            State([1, 1, 0, 1, 0, 0]),
            50000,
            post_selection=lambda s: s[1] == 1 and s[3] == 0,
        )
        results = backend.run(sampler)
        # Then add and re-sample
        circuit.herald((0, 1), 1)
        circuit.herald((2, 3), 0)
        sampler = Sampler(circuit, State([1, 1, 0, 0]), 50000)
        results2 = backend.run(sampler)
        # Check all results with outputs greater than 2000
        for s, c in results2.items():
            if c > 2000:
                full_s = s[0:1] + State([1]) + s[1:2] + State([0]) + s[2:]
                # Check results are within 10%
                assert pytest.approx(results[full_s], 0.1) == c

    @pytest.mark.parametrize("n_output", [0, 1])
    @pytest.mark.flaky(reruns=3)
    def test_herald_equivalent_lossy_imperfect_source(self, backend, n_output):
        """
        Checks that results are equivalent if a herald is used vs
        post-selection on a non-heralded lossy circuit, while also including an
        imperfect photon source.
        """
        circuit = Unitary(random_unitary(6))
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        # Define source to use
        source = Source(purity=0.9, brightness=0.9, indistinguishability=0.9)
        # Sampler without built-in heralds
        post_select = PostSelection()
        post_select.add(1, 1)
        post_select.add(3, n_output)
        sampler = Sampler(
            circuit,
            State([1, 1, 0, 1, 0, 0]),
            50000,
            source=source,
            post_selection=post_select,
        )
        results = backend.run(sampler)
        # Then add and re-sample
        circuit.herald((0, 1), 1)
        circuit.herald((2, 3), (0, n_output))
        sampler = Sampler(circuit, State([1, 1, 0, 0]), 50000, source=source)
        results2 = backend.run(sampler)
        # Check all results with outputs greater than 2000
        for s, c in results2.items():
            if c > 2000:
                full_s = (
                    s[0:1] + State([1]) + s[1:2] + State([n_output]) + s[2:]
                )
                # Check results are within 10%
                assert pytest.approx(results[full_s], 0.1) == c

    def test_hom_imperfect_brightness(self, backend):
        """
        Checks sampling a basic 2 photon input onto a 50:50 beam splitter,
        which should undergo HOM, producing outputs of |2,0> and |0,2>.
        Includes imperfect brightness.
        """
        circuit = PhotonicCircuit(2)
        circuit.bs(0)
        n_sample = 100000
        sampler = Sampler(
            circuit,
            State([1, 1]),
            n_sample,
            source=Source(brightness=0.8),
            random_seed=21,
            min_detection=2,
        )
        results = backend.run(sampler)
        assert len(results) == 2
        assert 0.49 < results[State([2, 0])] / n_sample < 0.51
        assert 0.49 < results[State([0, 2])] / n_sample < 0.51

    def test_loss_variable_value(self, backend):
        """
        Checks that Sampler is able to support number of required loss elements
        changing if these are part of a parameterized circuits.
        """
        loss = Parameter(0)
        circuit = PhotonicCircuit(4)
        circuit.bs(0, loss=loss)
        circuit.bs(2, loss=loss)
        circuit.bs(1, loss=loss)
        # Initially sample
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 10000)
        backend.run(sampler)
        # Add loss and resample
        loss.set(1)
        backend.run(sampler)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from random import choice, randint, random, sample

import pytest
from qiskit import QuantumCircuit
from qiskit.circuit.library import MCXGate

from lightworks import Sampler, Simulator, State, qubit
from lightworks.emulator import Backend
from lightworks.qubit import qiskit_converter
from lightworks.qubit.converter.converter import (
    ROTATION_GATES_MAP,
    SINGLE_QUBIT_GATES_MAP,
    THREE_QUBIT_GATES_MAP,
    TWO_QUBIT_GATES_MAP,
)
from lightworks.qubit.converter.utils import (
    convert_two_qubits_to_adjacent,
    post_selection_analyzer,
)
from lightworks.sdk.circuit.photonic_components import Barrier, Group

BACKEND = Backend("permanent")


class TestQiskitConversion:
    """
    Unit tests to check correct functionality of qiskit conversion function.
    """

    @pytest.mark.parametrize("gate", list(SINGLE_QUBIT_GATES_MAP.keys()))
    def test_all_single_qubit_gates(self, gate):
        """
        Checks all expected single qubit gates can be converted.
        """
        circ = QuantumCircuit(1)
        getattr(circ, gate)(0)
        qiskit_converter(circ)

    @pytest.mark.parametrize("gate", list(ROTATION_GATES_MAP.keys()))
    def test_all_rotation_qubit_gates(self, gate):
        """
        Checks all expected rotation qubit gates can be converted.
        """
        circ = QuantumCircuit(1)
        getattr(circ, gate)(math.tau * random(), 0)
        qiskit_converter(circ)

    def test_single_qubit_unitary(self):
        """
        Checks that a single qubit rotation gate can successfully be converted
        from qiskit to Lightworks.
        """
        circ = QuantumCircuit(1)
        circ.u(math.tau * random(), math.tau * random(), math.tau * random(), 0)
        qiskit_converter(circ)

    @pytest.mark.parametrize("gate", list(TWO_QUBIT_GATES_MAP.keys()))
    def test_all_two_qubit_gates(self, gate):
        """
        Checks all expected two qubit gates can be converted.
        """
        circ = QuantumCircuit(2)
        getattr(circ, gate)(0, 1)
        qiskit_converter(circ)

    @pytest.mark.parametrize("gate", list(THREE_QUBIT_GATES_MAP.keys()))
    def test_all_three_qubit_gates(self, gate):
        """
        Checks all expected three qubit gates can be converted.
        """
        circ = QuantumCircuit(3)
        getattr(circ, gate)(0, 1, 2)
        qiskit_converter(circ, allow_post_selection=True)

    @pytest.mark.filterwarnings("ignore::PendingDeprecationWarning")
    def test_four_qubit_gate(self):
        """
        Checks that an error is raised for a 4 qubit gate.
        """
        circ = QuantumCircuit(4)
        circ.append(MCXGate(3), [0, 1, 2, 3])
        with pytest.raises(ValueError):
            qiskit_converter(circ)

    def test_barrier(self):
        """
        Checks that barrier can be converted to lightworks.
        """
        circ = QuantumCircuit(3)
        circ.barrier([0, 1, 2])
        conv_circ = qiskit_converter(circ)[0]
        assert isinstance(conv_circ._get_circuit_spec()[0], Barrier)

    def test_x_gate(self):
        """
        Checks correct operation of a single qubit circuit with an X gate,
        flipping the qubit from the 0 to 1 state.
        """
        circ = QuantumCircuit(1)
        circ.x(0)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(conv_circ, State([1, 0]), State([0, 1]))
        results = BACKEND.run(sim)
        assert abs(results[State([1, 0]), State([0, 1])]) ** 2 == pytest.approx(
            1.0, 1e-6
        )

    def test_cnot(self):
        """
        Checks operation of two qubit CNOT gate produces output as expected.
        """
        circ = QuantumCircuit(2)
        circ.cx(0, 1)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(conv_circ, State([0, 1, 1, 0]), State([0, 1, 0, 1]))
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 1, 0]), State([0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 16, 1e-6)

    def test_cascaded_cnot(self):
        """
        Checks operation of two cascaded qubit CNOT gate produces output as
        expected.
        """
        circ = QuantumCircuit(3)
        circ.cx(0, 1)
        circ.cx(1, 2)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(
            conv_circ, State([0, 1, 1, 0, 1, 0]), State([0, 1, 0, 1, 0, 1])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 1, 0, 1, 0]), State([0, 1, 0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 16**2, 1e-6)

    def test_cnot_post_selected(self):
        """
        Checks operation of two qubit post-selected CNOT gate produces output as
        expected.
        """
        circ = QuantumCircuit(2)
        circ.cx(0, 1)
        conv_circ, _ = qiskit_converter(circ, allow_post_selection=True)

        sim = Simulator(conv_circ, State([0, 1, 1, 0]), State([0, 1, 0, 1]))
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 1, 0]), State([0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 9, 1e-6)

    def test_cascaded_cnot_post_selected(self):
        """
        Checks operation of two cascaded qubit post-selected CNOT gate produces
        output as expected. Both gates should be post-selected, so the success
        probability should be 1/9*1/9 = 1/81.
        """
        circ = QuantumCircuit(3)
        circ.cx(0, 1)
        circ.cx(1, 2)
        conv_circ, _ = qiskit_converter(circ, allow_post_selection=True)

        sim = Simulator(
            conv_circ, State([0, 1, 1, 0, 1, 0]), State([0, 1, 0, 1, 0, 1])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 1, 0, 1, 0]), State([0, 1, 0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 81, 1e-6)

    def test_bell_state(self):
        """
        Checks operation of two qubit H + CNOT gate produces the expected Bell
        state.
        """
        circ = QuantumCircuit(2)
        circ.h(0)
        circ.cx(0, 1)
        conv_circ, _ = qiskit_converter(circ)

        outputs = [State([1, 0, 1, 0]), State([0, 1, 0, 1])]
        sim = Simulator(conv_circ, State([1, 0, 1, 0]), outputs)
        results = BACKEND.run(sim)
        for out in outputs:
            assert abs(results[State([1, 0, 1, 0]), out]) ** 2 == pytest.approx(
                1 / 32, 1e-6
            )

    def test_cnot_flipped(self):
        """
        Checks operation of two qubit CNOT gate produces output as expected when
        the control and the target qubits are flipped.
        """
        circ = QuantumCircuit(2)
        circ.cx(1, 0)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(conv_circ, State([1, 0, 0, 1]), State([0, 1, 0, 1]))
        results = BACKEND.run(sim)
        assert abs(
            results[State([1, 0, 0, 1]), State([0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 16, 1e-6)

    def test_cnot_non_adjacent(self):
        """
        Checks operation of CNOT gate when the two qubits it is applied are not
        adjacent to each other.
        """
        circ = QuantumCircuit(3)
        circ.cx(0, 2)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(
            conv_circ, State([0, 1, 0, 1, 0, 1]), State([0, 1, 0, 1, 1, 0])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 0, 1, 0, 1]), State([0, 1, 0, 1, 1, 0])]
        ) ** 2 == pytest.approx(1 / 16, 1e-6)

    def test_cnot_non_adjacent_flipped(self):
        """
        Checks operation of CNOT gate when the two qubits it is applied are not
        adjacent to each other and the control and target are flipped.
        """
        circ = QuantumCircuit(3)
        circ.cx(2, 0)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(
            conv_circ, State([0, 1, 0, 1, 0, 1]), State([1, 0, 0, 1, 0, 1])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 0, 1, 0, 1]), State([1, 0, 0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 16, 1e-6)

    def test_ccnot(self):
        """
        Checks operation of three qubit CCNOT gate produces output as expected.
        """
        circ = QuantumCircuit(3)
        circ.ccx(0, 1, 2)
        conv_circ, _ = qiskit_converter(circ, allow_post_selection=True)

        sim = Simulator(
            conv_circ, State([0, 1, 0, 1, 1, 0]), State([0, 1, 0, 1, 0, 1])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 0, 1, 1, 0]), State([0, 1, 0, 1, 0, 1])]
        ) ** 2 == pytest.approx(1 / 72, 1e-6)

    def test_swap(self):
        """
        Checks swap is able to move correctly map between modes.
        """
        circ = QuantumCircuit(3)
        circ.swap(0, 2)
        conv_circ, _ = qiskit_converter(circ)

        sim = Simulator(
            conv_circ, State([0, 1, 0, 0, 2, 3]), State([2, 3, 0, 0, 0, 1])
        )
        results = BACKEND.run(sim)
        assert abs(
            results[State([0, 1, 0, 0, 2, 3]), State([2, 3, 0, 0, 0, 1])]
        ) ** 2 == pytest.approx(1, 1e-6)

    def test_complex_converter(self):
        """
        Tests conversion of a more complex qiskit circuit.
        """
        circ = QuantumCircuit(4)
        circ.x(0)
        circ.cx(0, 1)
        circ.y(0)
        circ.h(1)
        circ.cx(1, 2)
        circ.y(0)
        circ.z(2)
        circ.h(1)
        circ.cx(2, 3)
        qiskit_converter(circ, allow_post_selection=True)

    def test_post_selection(self):
        """
        Checks that post-selection object returned by the converter is able to
        create the required transformation.
        """
        circ = QuantumCircuit(2)
        circ.cx(0, 1)
        conv_circ, post_select = qiskit_converter(
            circ, allow_post_selection=True
        )
        # Run sampler and check results
        n_samples = 10000
        sampler = Sampler(
            conv_circ,
            State([0, 1, 1, 0]),
            n_samples,
            post_selection=post_select,
        )
        results = BACKEND.run(sampler)
        assert results[State([0, 1, 0, 1])] == n_samples

    def test_post_selection_rules(self):
        """
        Checks that post-selection objected returned by the converter contains
        rules on the correct modes.
        """
        circ = QuantumCircuit(2)
        circ.cx(0, 1)
        _, post_select = qiskit_converter(circ, allow_post_selection=True)

        r1_found = False
        r2_found = False
        for rule in post_select.rules:
            if rule.modes == (0, 1) and rule.n_photons == (1,):
                r1_found = True
            if rule.modes == (2, 3) and rule.n_photons == (1,):
                r2_found = True
        assert r1_found
        assert r2_found

    @pytest.mark.parametrize(
        ("q0", "q1"), [(0, 1), (1, 0), (2, 4), (4, 2), (2, 5), (5, 2)]
    )
    def test_convert_two_qubits_to_adjacent(self, q0, q1):
        """
        Checks that convert_two_qubits_to_adjacent function returns values with
        a difference of 1 and retains the correct order.
        """
        new_q0, new_q1, _ = convert_two_qubits_to_adjacent(q0, q1)
        assert abs(new_q1 - new_q0) == 1
        if q1 > q0:
            assert new_q1 > new_q0
        else:
            assert new_q1 < new_q0

    @pytest.mark.parametrize("n_qubits", [3, 4, 5])
    def test_post_selection_analyzer(self, n_qubits):
        """
        Checks that post-selection analyzer returns the correct number of
        elements.
        """
        circuit = build_random_qiskit_circuit(n_qubits)
        post_selects, _ = post_selection_analyzer(circuit)
        assert len(post_selects) == len(circuit.data)

    def test_single_qubit_rotation_rx_equivalance(self):
        """
        Checks that a qiskit single qubit rotation gate with correct angles
        successfully replicates the Rx gate in lightworks.
        """
        # Build circuit
        theta = math.tau * random()
        qc = QuantumCircuit(1)
        qc.u(theta, -math.pi / 2, math.pi / 2, 0)
        # Then convert to lightworks
        circ = qiskit_converter(qc)[0]
        # And check equivalence to Rx gate
        assert pytest.approx(qubit.Rx(theta).U) == circ.U

    def test_single_qubit_rotation_ry_equivalance(self):
        """
        Checks that a qiskit single qubit rotation gate with correct angles
        successfully replicates the Ry gate in lightworks.
        """
        # Build circuit
        theta = math.tau * random()
        qc = QuantumCircuit(1)
        qc.u(theta, 0, 0, 0)
        # Then convert to lightworks
        circ = qiskit_converter(qc)[0]
        # And check equivalence to Ry gate
        assert pytest.approx(qubit.Ry(theta).U) == circ.U

    def test_barrier_does_not_affect_post_selection(self):
        """
        Checks that barriers don't affect post-selection being added to a
        circuit.
        """
        # Build circuit
        qc = QuantumCircuit(2)
        qc.barrier()
        qc.cx(0, 1)
        qc.barrier()
        # Then convert to lightworks
        circ = qiskit_converter(qc)[0]
        # Check the relevant entry in the circuit spec is a heralded CNOT first
        assert isinstance(circ._get_circuit_spec()[1], Group)
        assert "Heralded" in circ._get_circuit_spec()[1].name
        # Then convert again with post-selection and check this is no longer the
        # case.
        circ = qiskit_converter(qc, allow_post_selection=True)[0]
        assert "Heralded" not in circ._get_circuit_spec()[1].name


def build_random_qiskit_circuit(n_qubits):
    """
    Builds a random qiskit circuit for testing
    """
    if n_qubits < 2:
        raise ValueError("Number of qubits should be at least two.")
    gates = ["x", "y", "z", "cx", "cz"]
    if n_qubits >= 3:
        gates += ["ccx", "ccz"]
    circuit = QuantumCircuit(n_qubits)
    for _i in range(randint(10, 20)):
        gate = choice(gates)
        # Create unique list of qubits for each gate
        qubits = sample(range(n_qubits), len(gate))
        getattr(circuit, gate)(*qubits)
    return circuit



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.state import State

from .mappings import PAULI_MAPPING, RHO_MAPPING
from .process_tomography import _ProcessTomography
from .projection import _vec
from .utils import _check_target_process, _combine_all


class GateFidelity(_ProcessTomography):
    """
    Computes the average gate fidelity using equation 19 from
    https://arxiv.org/pdf/quant-ph/0205035, which involves performing state
    tomography for a set of inputs.

    Args:

        n_qubits (int) : The number of qubits that will be used as part of the
            tomography.

        base_circuit (PhotonicCircuit) : An initial circuit which realises the
            required operation and can be modified for performing tomography.
            It is required that the number of circuit input modes equals 2 * the
            number of qubits.

    """

    @property
    def fidelity(self) -> float:
        """Returns the calculated fidelity of a process."""
        if not hasattr(self, "_fidelity"):
            raise AttributeError(
                "Fidelity has not yet been calculated, this can be achieved "
                "with the process method."
            )
        return self._fidelity

    def process(
        self,
        data: list[dict[State, int]] | dict[tuple[str, str], dict[State, int]],
        target_process: NDArray[np.complex128],
        project_to_physical: bool = False,
    ) -> float:
        """
        Calculates gate fidelity for an expected target process

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be tuple of the input and
                measurement basis.

            target_process (np.ndarray) : The unitary matrix corresponding to
                the target process. The dimension of this should be 2^n_qubits.

            project_to_physical (bool) : Controls whether the calculated density
                matrices are projected to a physical space. Defaults to False.

        Returns:

            float : The calculated fidelity.

        """
        target_process = _check_target_process(target_process, self.n_qubits)
        # Get density matrices and convert to vector
        rhos = self.process_density_matrices(data, project_to_physical)
        rho_vec = [rhos[i] for i in self._full_input_basis()]
        # Get pauli matrix basis and coefficients relating to unitary
        alpha_mat, u_basis = self._calculate_alpha_and_u_basis()
        # Find sum from equation
        total = 0
        for i, uj in enumerate(u_basis):
            for j, rho in enumerate(rho_vec):
                total += alpha_mat[i][j] * np.trace(
                    target_process
                    @ np.conj(uj.T)
                    @ np.conj(target_process.T)
                    @ rho
                )
        # Use total within calculation and return
        dim = 2**self.n_qubits
        self._fidelity = np.real((total + dim**2) / (dim**2 * (dim + 1))).item()
        return self.fidelity

    def _calculate_alpha_and_u_basis(
        self,
    ) -> tuple[NDArray[np.complex128], list[NDArray[np.complex128]]]:
        """
        Calculates the pauli matrix basis to use for the calculation and finds
        the coefficients of the alpha matrix used which can be used to the
        relate the input density matrices to the pauli matrix basis.
        """
        # Unitary basis
        u_basis = list(
            _combine_all(dict(PAULI_MAPPING), self.n_qubits).values()
        )
        # Input density matrices
        rho_basis = _combine_all(
            [RHO_MAPPING[i] for i in list(self._tomo_inputs)], self.n_qubits
        )
        # Then find alpha matrix
        alpha = np.zeros(
            (2 ** (2 * self.n_qubits), 2 ** (2 * self.n_qubits)), dtype=complex
        )
        basis_vectors = np.column_stack([_vec(m) for m in rho_basis])
        for i, u in enumerate(u_basis):
            alpha[i, :] = np.linalg.lstsq(basis_vectors, _vec(u))[0]
        return alpha, u_basis



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from collections import UserList
from dataclasses import dataclass

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State


@dataclass(slots=True, kw_only=True)
class _TomographyExperiment:
    """
    Contains the data for running a required state tomography experiment.
    """


@dataclass(slots=True, kw_only=True)
class StateTomographyExperiment(_TomographyExperiment):
    """
    Contains the data for running a required state tomography experiment.
    """

    circuit: PhotonicCircuit
    measurement_basis: str


@dataclass(slots=True, kw_only=True)
class ProcessTomographyExperiment(_TomographyExperiment):
    """
    Contains the data for running a required tomography experiment.
    """

    circuit: PhotonicCircuit
    input_state: State
    input_basis: str
    measurement_basis: str


TE = typing.TypeVar(
    "TE", StateTomographyExperiment, ProcessTomographyExperiment
)


class _TomographyList(UserList[TE]):
    """
    Base class for all list of tomography experiments.
    """

    @property
    def all_circuits(self) -> list[PhotonicCircuit]:
        """
        Returns a list of circuits corresponding to each of the tomography
        experiments in the list.
        """
        return [exp.circuit for exp in self]

    @property
    def all_measurement_basis(self) -> list[str]:
        """
        Returns a list of the measurement basis used for each tomography
        experiment.
        """
        return [exp.measurement_basis for exp in self]


class StateTomographyList(_TomographyList[StateTomographyExperiment]):
    """
    Stores a list of state tomography experiments.
    """


class ProcessTomographyList(_TomographyList[ProcessTomographyExperiment]):
    """
    Stores a list of tomography experiments.
    """

    @property
    def all_inputs(self) -> list[State]:
        """
        Returns a list of inputs corresponding to each of the tomography
        experiments in the list.
        """
        return [exp.input_state for exp in self]

    @property
    def all_input_basis(self) -> list[str]:
        """
        Returns a list of the input basis used for each tomography experiment.
        """
        return [exp.input_basis for exp in self]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from lightworks import Sampler, Simulator, State, Unitary, random_unitary
from lightworks.emulator import Backend, BackendError
from lightworks.emulator.backends import (
    PermanentBackend,
    SLOSBackend,
)


class TestBackend:
    """
    Unit tests for ensuring backend object remains functioning correctly.
    """

    @pytest.mark.parametrize("backend", ["permanent", "slos"])
    def test_valid_backend(self, backend):
        """Checks all currently valid backends can be set."""
        Backend(backend)

    def test_invalid_backends(self):
        """Checks an invalid backend raises a ValueError"""
        with pytest.raises(ValueError):
            Backend("not_a_backend")

    def test_invalid_backend_for_task(self):
        """
        Checks an error is raised when an invalid backend is chosen for a task.
        """
        backend = Backend("slos")
        sim = Simulator(Unitary(random_unitary(4)), State([1, 0, 1, 0]))
        with pytest.raises(BackendError):
            backend.run(sim)

    @pytest.mark.parametrize("backend_type", ["permanent", "slos"])
    def test_full_probability_distribution(self, backend_type):
        """
        Check against a known result for the full probability distribution
        while using the both permanent and slos backends.
        """
        backend = Backend(backend_type)._backend
        unitary = random_unitary(3, seed=2)
        dist = backend.full_probability_distribution(
            Unitary(unitary)._build(), State([1, 1, 0])
        )
        known_dist = {
            State([2, 0, 0]): 0.2600968016733883,
            State([1, 1, 0]): 0.22397644081164791,
            State([0, 2, 0]): 0.13677578896115924,
            State([1, 0, 1]): 0.04763948802896683,
            State([0, 1, 1]): 0.2377381500275615,
            State([0, 0, 2]): 0.09377333049727585,
        }
        for d in dist:
            if d not in known_dist:
                pytest.fail(
                    "State from produced distribution not in known "
                    "distribution."
                )
            assert known_dist[d] == pytest.approx(dist[d], 1e-8)

    def test_permanent_slos_equivalence(self):
        """
        Check permanent and slos backends produce equivalent probability
        distributions.
        """
        circuit = Unitary(random_unitary(6))
        input_state = State([1, 0, 1, 0, 1, 0])
        # Find distributions
        backend_p = PermanentBackend()
        p1 = backend_p.full_probability_distribution(
            circuit._build(), input_state
        )
        backend_s = SLOSBackend()
        p2 = backend_s.full_probability_distribution(
            circuit._build(), input_state
        )
        # Check equivalence
        for s in p1:
            if round(p1[s], 8) != round(p2[s], 8):
                pytest.fail("Methods do not produce equivalent distributions.")

    def test_backend_str_return(self):
        """
        Check that backend value is stored and returned correctly when using
        the str operator.
        """
        backend = Backend("permanent")
        assert str(backend) == "permanent"

    def test_backend_repr_return(self):
        """
        Checks that backend value is correctly included in __repr__ for
        backend.
        """
        backend = Backend("permanent")
        assert "permanent" in repr(backend)

    def test_backend_call(self):
        """
        Checks that a backend can be called directly by providing a target task.
        """
        backend = Backend("permanent")
        sim = Simulator(Unitary(random_unitary(4)), State([1, 0, 1, 0]))
        backend(sim)

    def test_backend_call_equivalance(self):
        """
        Checks that a backend can be called directly and this produces
        equivalent results to run.
        """
        backend = Backend("permanent")
        sim = Simulator(Unitary(random_unitary(4)), State([1, 0, 1, 0]))
        assert backend.run(sim) == backend(sim)


class TestPermanent:
    """
    Specific functions for testing Permanent calculation remains functional
    and consistent.
    """

    def test_single_photon_return(self):
        """
        Confirm elements of unitary are returned by permanent in single photon
        cases.
        """
        unitary = random_unitary(4)
        # Diagonal
        assert unitary[0, 0] == PermanentBackend().probability_amplitude(
            unitary, State([1, 0, 0, 0]), State([1, 0, 0, 0])
        )
        # Off-diagonal
        assert unitary[2, 1] == PermanentBackend().probability_amplitude(
            unitary, State([0, 1, 0, 0]), State([0, 0, 1, 0])
        )

    def test_known_result(self):
        """
        Calculate a permanent value and check it matches a previously
        calculated result.
        """
        unitary = random_unitary(6, seed=23)
        r = PermanentBackend().probability_amplitude(
            unitary, State([1, 0, 1, 0, 1, 0]), State([0, 1, 1, 0, 0, 1])
        )
        assert r == pytest.approx(-0.02042658999324299 - 0.02226528732909283j)

    def test_probability_amplitude(self):
        """
        Confirms that expected elements from the unitary matrix are returned
        for a single photon input.
        """
        backend = PermanentBackend()
        unitary = random_unitary(4)
        # Diagonal
        assert unitary[0, 0] == backend.probability_amplitude(
            unitary, [1, 0, 0, 0], [1, 0, 0, 0]
        )
        # Off-diagonal
        assert unitary[2, 1] == backend.probability_amplitude(
            unitary, [0, 1, 0, 0], [0, 0, 1, 0]
        )

    def test_probability(self):
        """
        Confirms that expected probability is returned for a single photon
        input.
        """
        backend = PermanentBackend()
        unitary = random_unitary(4)
        # Diagonal
        assert abs(unitary[0, 0]) ** 2 == backend.probability(
            unitary, [1, 0, 0, 0], [1, 0, 0, 0]
        )
        # Off-diagonal
        assert abs(unitary[2, 1]) ** 2 == backend.probability(
            unitary, [0, 1, 0, 0], [0, 0, 1, 0]
        )

    def test_probability_amplitude_multi(self):
        """
        Confirms that expected elements from the unitary matrix are returned
        for a multi photon input.
        """
        backend = PermanentBackend()
        unitary = random_unitary(4, seed=10)
        # Diagonal
        pa = backend.probability_amplitude(unitary, [1, 0, 0, 0], [1, 0, 0, 0])
        assert pa == pytest.approx(0.429095917729817 - 0.366263376556379j, 1e-8)
        # Off-diagonal
        pa = backend.probability_amplitude(unitary, [0, 1, 0, 0], [0, 0, 1, 0])
        assert pa == pytest.approx(
            -0.15003076436547 + 0.4696358907386921j, 1e-8
        )

    def test_probability_multi(self):
        """
        Confirms that expected probability is returned for a multi photon
        input.
        """
        backend = PermanentBackend()
        unitary = random_unitary(4, seed=11)
        # Diagonal
        p = backend.probability(unitary, [1, 0, 0, 0], [1, 0, 0, 0])
        assert p == pytest.approx(0.6122546643219795, 1e-8)
        # Off-diagonal
        p = backend.probability(unitary, [0, 1, 0, 0], [0, 0, 1, 0])
        assert p == pytest.approx(0.25051188442720407, 1e-8)


class TestSlos:
    """
    Specific tests for SLOS function to check for errors in calculation.
    """

    def test_hom(self):
        """Check hom result and ensure returned value is as expected."""
        unitary = np.array([[1, 1j], [1j, 1]]) * 1 / (2**0.5)
        r = SLOSBackend().calculate(unitary, State([1, 1]))
        assert r[1, 1] == 0
        assert r[2, 0] == pytest.approx(0.7071067811865475j)

    def test_known_result(self):
        """
        Check produced output matches a previously calculated result.
        """
        unitary = random_unitary(6, seed=23)
        r = SLOSBackend().calculate(unitary, State([1, 0, 1, 0, 1, 0]))
        assert r[1, 1, 1, 0, 0, 0] == pytest.approx(
            -0.0825807219472892 + 0.0727188703263498j
        )


class TestFockBackend:
    """
    Unit tests for FockBackend baseclass.
    """

    @pytest.mark.parametrize("backend", [PermanentBackend(), SLOSBackend()])
    def test_sampler_results_cached(self, backend):
        """
        Confirms that cached results are used when the same Sampler task is run
        on the backend.
        """
        circuit = Unitary(random_unitary(4))
        # Run initial experiment
        sampler = Sampler(circuit, State([1, 0, 1, 0]), 1000)
        backend.run(sampler)
        # Get data and check if it matches the cache
        results = backend._check_cache(sampler._generate_task())
        assert results is not None



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.state import State

from .mappings import PAULI_MAPPING, RHO_MAPPING
from .process_tomography import _ProcessTomography
from .projection import _unvec, _vec, project_choi_to_physical
from .utils import (
    _calculate_expectation_value,
    _combine_all,
    _get_tomo_measurements,
    process_fidelity,
)


class MLEProcessTomography(_ProcessTomography):
    """
    Runs quantum process tomography using the maximum likelihood estimation
    method.

    Args:

        n_qubits (int) : The number of qubits that will be used as part of the
            tomography.

        base_circuit (PhotonicCircuit) : An initial circuit which realises the
            required operation and can be modified for performing tomography.
            It is required that the number of circuit input modes equals 2 * the
            number of qubits.

    """

    _tomo_inputs = ("Z+", "Z-", "X+", "X-", "Y+", "Y-")

    @property
    def choi(self) -> NDArray[np.complex128]:
        """Returns the calculate choi matrix for a circuit."""
        if not hasattr(self, "_choi"):
            raise AttributeError(
                "Choi matrix has not yet been calculated, this can be achieved "
                "with the process method."
            )
        return self._choi

    def process(
        self,
        data: list[dict[State, int]] | dict[tuple[str, str], dict[State, int]],
    ) -> NDArray[np.complex128]:
        """
        Performs process tomography with the configured elements and calculates
        the choi matrix using maximum likelihood estimation.

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be tuple of the input and
                measurement basis.

        Returns:

            np.ndarray : The calculated choi matrix for the process.

        """
        results = self._convert_tomography_data(data)
        nij = {}
        # Convert counts to an expectation value
        for (in_state, meas), result in results.items():
            # Remove trivial measurement here
            if meas == ",".join("I" * self.n_qubits):
                continue
            nij[in_state, meas] = _calculate_expectation_value(meas, result)
        # Run MLE algorithm and get choi
        mle = MLETomographyAlgorithm(self.n_qubits, self._full_input_basis())
        self._choi = mle.pgdb(nij)
        return self.choi

    def fidelity(self, choi_exp: NDArray[np.complex128]) -> float:
        """
        Calculates fidelity of the calculated choi matrix compared to the
        expected one.
        """
        return process_fidelity(self.choi, choi_exp)


class MLETomographyAlgorithm:
    """
    Implements the pgdB algorithm for maximum likelihood estimation from
    https://arxiv.org/abs/1803.10062 for calculation of a physical choi matrix
    from the tomography measurement data.

    Args:

        n_qubits (int) : The number of qubits used within the tomography.

    """

    def __init__(self, n_qubits: int, input_basis: list[str]) -> None:
        self.n_qubits = n_qubits

        # Pre-calculate all required n_qubit pauli & density matrices + the
        # inputs and measurement basis for the tomography
        self._all_rhos = _combine_all(RHO_MAPPING, self.n_qubits)
        self._all_pauli = _combine_all(PAULI_MAPPING, self.n_qubits)
        self._input_basis = list(input_basis)
        self._meas_basis = _get_tomo_measurements(n_qubits, remove_trivial=True)

        self._a_matrix = self._a_mat()

    def pgdb(
        self,
        data: dict[tuple[str, str], float],
        max_iter: int = 1000,
        stop_threshold: float = 1e-10,
    ) -> NDArray[np.complex128]:
        """
        Runs the pgdB algorithm on the provided data set.

        Args:

            data (dict) : The measured tomography experiment data. The keys of
                this dictionary should be the input/measurement basis and the
                values should be the calculated expectation values.

            max_iter (int, optional) : Sets the maximum number of iterations
                that the algorithm can perform, defaults to 1000.

            stop_threshold (float, optional) : Sets the stopping threshold for
                the gradient descent algorithm. Defaults to 1e-10.

        Returns:

            np.ndarray : The calculated choi matrix

        """
        # Convert data to required vector format
        n_vec = self._n_vec_from_data(data)

        # Initialize choi as identity and find initial cost
        dim = 2**self.n_qubits
        choi = np.identity(dim**2, dtype=complex) / dim
        current_cost = self._cost(choi, n_vec)

        # Define algorithm parameters
        mu = 3 / (2 * dim**2)
        gamma = 0.3

        # Run gradient descent
        for _ in range(max_iter):
            # Find modification to the choi matrix
            mod = project_choi_to_physical(
                choi - 1 / mu * self._gradient(choi, n_vec),
                max_iter=max_iter,
            )
            mod -= choi

            # Optimise alpha weighting parameter
            alpha = 0.5
            new_cost = self._cost(choi + alpha * mod, n_vec)
            thresh_value = gamma * np.trace(
                mod @ np.conj(self._gradient(choi.T, n_vec))
            )
            while new_cost > current_cost + alpha * thresh_value:
                alpha *= 0.5
                new_cost = self._cost(choi + alpha * mod, n_vec)

            # Update choi
            choi += alpha * mod

            # Check cost is still improving, otherwise stop
            if current_cost - new_cost < stop_threshold:
                break
            current_cost = new_cost
        # Warn if the max iterations are exceeded
        else:
            warnings.warn("Max iterations exceeded.", stacklevel=1)

        return choi

    def _n_vec_from_data(
        self, data: dict[tuple[str, str], float]
    ) -> NDArray[np.float64]:
        """
        Converts the data from dictionary format into the required vector,
        ensuring this is correct regardless of the data ordering of the
        dictionary.
        """
        n_vec = np.zeros(2 * len(data), dtype=complex)
        for i, in_s in enumerate(self._input_basis):
            for j, meas in enumerate(self._meas_basis):
                coord = 2 * (len(self._meas_basis) * i + j)
                n = data[in_s, meas]
                n_vec[coord : coord + 2] = [(1 + n) / 2, (1 - n) / 2]
        return n_vec / len(data)

    def _a_mat(self) -> NDArray[np.complex128]:
        """
        Calculates the A matrix which is used for vectorisation of the pgdB
        algorithm.
        """
        dim1 = len(self._input_basis) * len(self._meas_basis) * 2
        dim2 = 2 ** (4 * self.n_qubits)
        a_mat = np.zeros((dim1, dim2), dtype=complex)
        id_mat = np.identity(2**self.n_qubits, dtype=complex)
        for i, in_s in enumerate(self._input_basis):
            for j, meas in enumerate(self._meas_basis):
                obs = self._all_pauli[meas]
                a_mat[2 * (len(self._meas_basis) * i + j), :] = _vec(
                    np.kron(self._all_rhos[in_s], ((id_mat + obs) / 2).T)
                )[:]
                a_mat[2 * (len(self._meas_basis) * i + j) + 1, :] = _vec(
                    np.kron(self._all_rhos[in_s], ((id_mat - obs) / 2).T)
                )[:]
        return a_mat / (2 ** (2 * self.n_qubits))

    def _p_vec(self, choi: NDArray[np.complex128]) -> NDArray[np.complex128]:
        """
        Calculates the expected measurement outcomes from the provided choi
        matrix.
        """
        return (self._a_matrix @ _vec(choi.T)).clip(1e-8)

    def _cost(
        self, choi: NDArray[np.complex128], n_vec: NDArray[np.float64]
    ) -> NDArray[np.complex128]:
        """
        Computes the variation between the current choi matrix and the
        calculated values.
        """
        return -n_vec.T @ np.log(self._p_vec(choi))

    def _gradient(
        self, choi: NDArray[np.complex128], n_vec: NDArray[np.float64]
    ) -> NDArray[np.complex128]:
        """
        Finds gradient between expected and measured expectation values.
        """
        return -_unvec(np.conj(self._a_matrix.T) @ (n_vec / self._p_vec(choi)))



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

import numpy as np
from numpy.typing import NDArray

from lightworks.emulator.utils.sim import check_photon_numbers
from lightworks.sdk.results import SimulationResult
from lightworks.sdk.state import State
from lightworks.sdk.tasks import AnalyzerTask
from lightworks.sdk.utils.exceptions import PhotonNumberError
from lightworks.sdk.utils.heralding import add_heralds_to_state
from lightworks.sdk.utils.post_selection import DefaultPostSelection

from .runner import RunnerABC


class AnalyzerRunner(RunnerABC):
    """
    Calculates an output probability distribution under the assigned set of
    post-selection + heralding constraints.

    Args:

        data (AnalyzerTask) : The task which is to be executed.

        probability_function (Callable) : Function for calculating probability
            of transition between an input and output for a given unitary.

        state_generator (Callable) : A function for generating the basis states
            for a given number of modes and photons.

    """

    def __init__(
        self,
        data: AnalyzerTask,
        probability_function: Callable[
            [NDArray[np.complex128], list[int], list[int]], float
        ],
        state_generator: Callable[[int, int], list[list[int]]],
    ) -> None:
        self.data = data
        self.probability_function = probability_function
        self.state_generator = state_generator

    def run(self) -> SimulationResult:
        """
        Runs analyzer process using the configured data.

        Returns:

            SimulationResult : A dictionary containing an array of probability
                values between the provided inputs/outputs.

        """
        # Process inputs, adding heralds and loss modes
        inputs = list(self.data.inputs)  # Copy input list
        check_photon_numbers(inputs, inputs[0].n_photons)
        in_heralds = self.data.circuit.heralds.input
        full_inputs = [
            add_heralds_to_state(i, in_heralds)
            + [0] * self.data.circuit.loss_modes
            for i in inputs
        ]
        n_photons = sum(full_inputs[0]) - sum(
            self.data.circuit.heralds.output.values()
        )
        # Generate lists of possible outputs with and without heralded modes
        full_outputs, filtered_outputs = self._generate_outputs(
            self.data.circuit.input_modes, n_photons
        )
        # Calculate permanent for the given inputs and outputs and return
        # values
        probs = self._get_probs(full_inputs, full_outputs)
        # Calculate performance by finding sum of valid transformations
        performance = probs.sum() / len(full_inputs)
        # Analyse error rate from expected results if specified
        if self.data.expected is not None:
            error_rate = self._calculate_error_rate(
                probs, inputs, filtered_outputs, self.data.expected
            )
        else:
            error_rate = None
        # Compile results into results object
        results = SimulationResult(
            probs,
            "probability",
            inputs=inputs,
            outputs=filtered_outputs,
            performance=performance,
        )
        if error_rate is not None:
            results.error_rate = error_rate  # type: ignore[attr-defined]
        # Return dict
        return results

    def _get_probs(
        self, full_inputs: list[list[int]], full_outputs: list[list[int]]
    ) -> NDArray[np.float64]:
        """
        Create an array of output probabilities for a given set of inputs and
        outputs.
        """
        probs = np.zeros((len(full_inputs), len(full_outputs)))
        for i, ins in enumerate(full_inputs):
            for j, outs in enumerate(full_outputs):
                # No loss case
                if not self.data.circuit.loss_modes:
                    probs[i, j] += self.probability_function(
                        self.data.circuit.U_full, ins, outs
                    )
                # Lossy case
                # Photon number preserved
                elif sum(ins) == sum(outs):
                    probs[i, j] += self.probability_function(
                        self.data.circuit.U_full,
                        ins,
                        outs + [0] * self.data.circuit.loss_modes,
                    )
                # Otherwise
                else:
                    # If n_out < n_in work out all loss mode combinations
                    # and find probability of each
                    n_loss = sum(ins) - sum(outs)
                    if n_loss < 0:
                        raise PhotonNumberError(
                            "Output photon number larger than input number."
                        )
                    # Find loss states and find probability of each
                    loss_states = self.state_generator(
                        self.data.circuit.loss_modes, n_loss
                    )
                    for ls in loss_states:
                        fs = outs + ls
                        probs[i, j] += self.probability_function(
                            self.data.circuit.U_full, ins, fs
                        )

        return probs

    def _calculate_error_rate(
        self,
        probabilities: NDArray[np.float64],
        inputs: list[State],
        outputs: list[State],
        expected: dict[State, State | list[State]],
    ) -> float:
        """
        Calculate the error rate for a set of expected mappings between inputs
        and outputs, given the results calculated by the analyzer.
        """
        # Check all inputs in expectation mapping
        for s in inputs:
            if s not in expected:
                msg = f"Input state {s} not in provided expectation dict."
                raise KeyError(msg)
        # For each input check error rate
        errors = []
        for i, s in enumerate(inputs):
            out = expected[s]
            # Convert expected output to list if only one value provided
            if isinstance(out, State):
                out = [out]
            iprobs = probabilities[i, :]
            error = 1
            # Loop over expected outputs and subtract from error value
            for o in out:
                if o in outputs:
                    loc = outputs.index(o)
                    error -= iprobs[loc] / sum(iprobs)
            errors += [error]
        # Then take average and return
        return float(np.mean(errors))

    def _generate_outputs(
        self, n_modes: int, n_photons: int
    ) -> tuple[list[list[int]], list[State]]:
        """
        Generates all possible outputs for a given number of modes, photons and
        heralding + post-selection conditions. It returns two list, one with
        the heralded modes included and one without.
        """
        # Get all possible outputs for the non-herald modes
        if not self.data.circuit.loss_modes:
            outputs = self.state_generator(n_modes, n_photons)
        # Combine all n < n_in for lossy case
        else:
            outputs = []
            for n in range(n_photons + 1):
                outputs += self.state_generator(n_modes, n)
        # Filter outputs according to post selection and add heralded photons
        filtered_outputs = []
        full_outputs = []
        out_heralds = self.data.circuit.heralds.output
        post_selection = (
            DefaultPostSelection()
            if self.data.post_selection is None
            else self.data.post_selection
        )
        for state in outputs:
            # Check output meets all post selection rules
            if post_selection.validate(state):
                filtered_outputs += [State(state)]
                full_outputs += [add_heralds_to_state(state, out_heralds)]
        # Check some valid outputs found
        if not full_outputs:
            raise ValueError(
                "No valid outputs found, consider relaxing post-selection."
            )

        return (full_outputs, filtered_outputs)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from lightworks.qubit.gates.single_qubit_gates import (
    SX,
    H,
    P,
    Rx,
    Ry,
    Rz,
    S,
    Sadj,
    T,
    Tadj,
    X,
    Y,
    Z,
)
from lightworks.qubit.gates.three_qubit_gates import CCNOT, CCZ
from lightworks.qubit.gates.two_qubit_gates import (
    CNOT,
    CZ,
    SWAP,
    CNOT_Heralded,
    CZ_Heralded,
)
from lightworks.sdk.circuit import PhotonicCircuit, Unitary

from .utils import convert_two_qubits_to_adjacent

SINGLE_QUBIT_GATES_MAP = {
    "h": H(),
    "x": X(),
    "y": Y(),
    "z": Z(),
    "s": S(),
    "sdg": Sadj(),
    "t": T(),
    "tdg": Tadj(),
    "sx": SX(),
}
ROTATION_GATES_MAP = {"rx": Rx, "ry": Ry, "rz": Rz, "p": P}

TWO_QUBIT_GATES_MAP = {"cx": CNOT_Heralded, "cz": CZ_Heralded, "swap": SWAP}
TWO_QUBIT_GATES_MAP_PS = {"cx": CNOT, "cz": CZ}

THREE_QUBIT_GATES_MAP = {"ccx": CCNOT, "ccz": CCZ}


class Converter:
    """
    TODO: DESC
    """

    def __init__(
        self, n_qubits: int, allow_post_selection: bool = False
    ) -> None:
        self.allow_post_selection = allow_post_selection
        self._modes = {i: (2 * i, 2 * i + 1) for i in range(n_qubits)}

    def _add_single_qubit_gate(
        self, circuit: PhotonicCircuit, gate: str, qubit: int
    ) -> None:
        """
        Adds a single qubit gate to the selected qubit on the circuit.
        """
        circuit.add(SINGLE_QUBIT_GATES_MAP[gate], self._modes[qubit][0])

    def _add_single_qubit_rotation_gate(
        self, circuit: PhotonicCircuit, gate: str, theta: float, qubit: int
    ) -> None:
        """
        Adds a single qubit gate to the selected qubit on the circuit.
        """
        circuit.add(ROTATION_GATES_MAP[gate](theta), self._modes[qubit][0])

    def _add_single_qubit_unitary(
        self, circuit: PhotonicCircuit, params: list[float], qubit: int
    ) -> None:
        """
        Adds an arbitrary single qubit rotation unitary to a circuit
        """
        if len(params) != 3:
            raise ValueError("Expected unitary gate to have 3 parameters.")
        theta, phi, lam = params
        unitary = np.array(
            [
                [np.cos(theta / 2), -np.exp(1j * lam) * np.sin(theta / 2)],
                [
                    np.exp(1j * phi) * np.sin(theta / 2),
                    np.exp(1j * (phi + lam)) * np.cos(theta / 2),
                ],
            ]
        )
        circuit.add(Unitary(unitary), self._modes[qubit][0])

    def _add_two_qubit_gate(
        self,
        circuit: PhotonicCircuit,
        gate: str,
        q0: int,
        q1: int,
        post_selection: bool = False,
    ) -> None:
        """
        Adds a two qubit gate to the circuit on the selected qubits.
        """
        if gate == "swap":
            circuit.add(
                TWO_QUBIT_GATES_MAP["swap"](self._modes[q0], self._modes[q1]), 0
            )
        elif gate in {"cx", "cz"}:
            mapper = (
                TWO_QUBIT_GATES_MAP
                if not post_selection
                else TWO_QUBIT_GATES_MAP_PS
            )
            q0, q1, to_swap = convert_two_qubits_to_adjacent(q0, q1)
            if gate == "cx":
                target = q1 - min([q0, q1])
                add_circ = mapper["cx"](target)
            else:
                add_circ = mapper["cz"]()
            add_mode = self._modes[min([q0, q1])][0]
            for swap_qs in to_swap:
                self._add_two_qubit_gate(
                    circuit, "swap", swap_qs[0], swap_qs[1]
                )
            circuit.add(add_circ, add_mode)
            for swap_qs in to_swap:
                self._add_two_qubit_gate(
                    circuit, "swap", swap_qs[0], swap_qs[1]
                )
        else:
            msg = f"Unsupported gate '{gate}' included in circuit."
            raise ValueError(msg)

    def _add_three_qubit_gate(
        self,
        circuit: PhotonicCircuit,
        gate: str,
        q0: int,
        q1: int,
        q2: int,
        post_selection: bool = False,
    ) -> None:
        """
        Adds a three qubit gate to the circuit on the selected qubits.
        """
        if gate in {"ccx", "ccz"}:
            if not post_selection:
                raise ValueError(
                    "Three qubit gates can only be used with post-selection. "
                    "Ensure allow_post_selection is True to enable this. The "
                    "location of the gate may also need to be towards the end "
                    "of the circuit as a result of the requirements on "
                    "post-selection."
                )
            all_qubits = [q0, q1, q2]
            if max(all_qubits) - min(all_qubits) != 2:
                raise ValueError(
                    "CCX and CCZ qubits must be adjacent to each other, "
                    "please add swap gates to achieve this."
                )
            if gate == "ccx":
                target = q2 - min(all_qubits)
                add_circ = THREE_QUBIT_GATES_MAP["ccx"](target)
            else:
                add_circ = THREE_QUBIT_GATES_MAP["ccz"]()
            add_mode = self._modes[min(all_qubits)][0]
            circuit.add(add_circ, add_mode)
        else:
            msg = f"Unsupported gate '{gate}' included in circuit."
            raise ValueError(msg)

    def _add_barrier(self, circuit: PhotonicCircuit, qubits: list[int]) -> None:
        """
        Adds a barrier to the circuit on the provided qubits.
        """
        modes = [self._modes[q][i] for q in qubits for i in range(2)]
        circuit.barrier(modes)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.utils.exceptions import LightworksError
from lightworks.sdk.utils.post_selection import PostSelection

from . import QISKIT_INSTALLED
from .converter import (
    ROTATION_GATES_MAP,
    SINGLE_QUBIT_GATES_MAP,
    THREE_QUBIT_GATES_MAP,
    TWO_QUBIT_GATES_MAP,
    Converter,
)
from .utils import post_selection_analyzer

if QISKIT_INSTALLED:
    from qiskit import QuantumCircuit


ALLOWED_GATES = [
    *SINGLE_QUBIT_GATES_MAP,
    *ROTATION_GATES_MAP,
    *TWO_QUBIT_GATES_MAP,
    *THREE_QUBIT_GATES_MAP,
    "barrier",
    "u",
]


def qiskit_converter(
    circuit: "QuantumCircuit", allow_post_selection: bool = False
) -> tuple[PhotonicCircuit, PostSelection | None]:
    """
    Performs conversion of a provided qiskit QuantumCircuit into a photonic
    circuit within Lightworks.

    Args:

        circuit (QuantumCircuit) : The qiskit circuit to be converted.

        allow_post_selection (bool, optional) : Controls whether post-selected
            gates can be utilised within the circuit.

    Returns:

        PhotonicCircuit : The created circuit within Lightworks.

        PostSelection | None : If post-selection rules are required for the
            created circuit, then an object which implements these will be
            returned, otherwise it will be None.

    """
    converter = QiskitConverter(circuit.num_qubits, allow_post_selection)
    return converter.convert(circuit)


class QiskitConverter(Converter):
    """
    Manages conversion between qiskit and lightworks circuit, adding each of the
    qubit gates into a created circuit object.

    Args:

        allow_post_selection (bool, optional) : Controls whether post-selected
            gates can be utilised within the circuit.

    """

    def __init__(
        self, n_qubits: int, allow_post_selection: bool = False
    ) -> None:
        super().__init__(n_qubits, allow_post_selection)

    def convert(
        self, q_circuit: "QuantumCircuit"
    ) -> tuple[PhotonicCircuit, PostSelection | None]:
        """
        Performs conversion of a provided qiskit QuantumCircuit into a photonic
        circuit within Lightworks.

        Args:

            q_circuit (QuantumCircuit) : The qiskit circuit to be converted.

        Returns:

            PhotonicCircuit : The created circuit within Lightworks.

            PostSelection | None : If post-selection rules are required for the
                created circuit, then an object which implements these will be
                returned, otherwise it will be None.

        """
        if not QISKIT_INSTALLED:
            raise LightworksError(
                "Lightworks qiskit optional requirements not installed, "
                "this can be achieved with 'pip install lightworks[qiskit]'."
            )

        if not isinstance(q_circuit, QuantumCircuit):
            raise TypeError(
                "PhotonicCircuit to convert must be a qiskit circuit."
            )

        n_qubits = q_circuit.num_qubits
        circuit = PhotonicCircuit(n_qubits * 2)

        if self.allow_post_selection:
            post_select, ps_qubits = post_selection_analyzer(q_circuit)
        else:
            post_select = [False] * len(q_circuit.data)

        for i, inst in enumerate(q_circuit.data):
            gate = inst.operation.name
            qubits = [
                inst.qubits[i]._index for i in range(inst.operation.num_qubits)
            ]
            if gate not in ALLOWED_GATES:
                msg = f"Unsupported gate '{gate}' included in circuit."
                raise ValueError(msg)
            # First catch barriers
            if gate == "barrier":
                self._add_barrier(circuit, qubits)
            # Single Qubit Gates
            elif len(qubits) == 1:
                if gate == "u":
                    self._add_single_qubit_unitary(
                        circuit, inst.params, qubits[0]
                    )
                elif gate in SINGLE_QUBIT_GATES_MAP:
                    self._add_single_qubit_gate(circuit, gate, qubits[0])
                else:
                    theta = inst.operation.params[0]
                    self._add_single_qubit_rotation_gate(
                        circuit, gate, theta, qubits[0]
                    )
            # Two Qubit Gates
            elif len(qubits) == 2:
                self._add_two_qubit_gate(
                    circuit,
                    gate,
                    qubits[0],
                    qubits[1],
                    post_select[i],
                )
            # Three Qubit Gates
            elif len(qubits) == 3:
                self._add_three_qubit_gate(
                    circuit,
                    gate,
                    qubits[0],
                    qubits[1],
                    qubits[2],
                    post_select[i],
                )
            # Limit to three qubit gates
            else:
                raise ValueError("Gates with more than 3 qubits not supported.")

        ps_rules = None
        if self.allow_post_selection and ps_qubits:
            ps_rules = PostSelection()
            for q in ps_qubits:
                ps_rules.add(self._modes[q], 1)

        return (circuit, ps_rules)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lightworks Emulator
===================

This module is designed for the simulation of boson sampling in linear optic
photonic circuits. It contains a number of different simulation options, with
each intended for a different particular purpose. The module also contains
capability to simulate a number of imperfections which are typically present
in a boson sampling experiment, such as loss and an imperfect photon source or
detectors.

Simulators:

    Simulator : Directly calculates the probability amplitudes from transitions
        between given inputs and outputs on a circuit.

    Sampler : Calculates the output distribution for a given input state and
        circuit, and enables sampling from it. Imperfect sources and detectors
        can also be utilised here.

    Analyzer : Finds all possible outputs of a circuit with a given set of
        inputs, conditional on them meeting a set of post-selection and
        heralding criteria. This means it can be used to analyze how well a
        circuit performs for a given task. It can also produce an error rate
        and value for circuit performance (how often a valid output will be
        produced).

"""

from .backends import Backend
from .components import Detector, Source
from .utils.exceptions import *  # noqa: F403

__all__ = [
    "Backend",
    "Detector",
    "Source",
]



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from typing import Any

from multimethod import multimethod

from lightworks.emulator.state import AnnotatedState
from lightworks.sdk.circuit.photonic_compiler import CompiledPhotonicCircuit
from lightworks.sdk.state import State


@multimethod
def pdist_calc(
    circuit: CompiledPhotonicCircuit,
    inputs: dict[State, int | float],
    probability_func: Callable[
        [CompiledPhotonicCircuit, State], dict[State, float]
    ],
) -> dict[State, float]:
    """
    Calculate the output state probability distribution for cases where
    inputs are state objects. This is the case when the source is perfect
    or only an imperfect brightness is used.

    Args:

        circuit (CompiledPhotonicCircuit) : The compiled circuit that is to be
            sampled from.

        inputs (dict) : The inputs to the system and their associated
            probabilities.

        probability_func (Callable) : A method for calculation of a probability
            distribution, given a circuit and input state.

    Returns:

        dict : The calculated output probability distribution.

    """
    pdist: dict[State, float] = {}
    # Loop over each possible input
    for istate, prob in inputs.items():
        # Calculate sub distribution
        sub_dist = probability_func(circuit, istate)
        if not pdist:
            if prob == 1:
                pdist = sub_dist
            else:
                pdist = {s: p * prob for s, p in sub_dist.items()}
        else:
            for s, p in sub_dist.items():
                if s in pdist:
                    pdist[s] += p * prob
                else:
                    pdist[s] = p * prob
    # Calculate zero photon state probability afterwards
    total_prob = sum(pdist.values())
    if total_prob < 1 and circuit.loss_modes > 0:
        pdist[State([0] * circuit.n_modes)] = 1 - total_prob

    return pdist


@pdist_calc.register
def annotated_state_pdist_calc(
    circuit: CompiledPhotonicCircuit,
    inputs: dict[AnnotatedState, int | float],
    probability_func: Callable[
        [CompiledPhotonicCircuit, State], dict[State, float]
    ],
) -> dict[State, float]:
    """
    Perform output state probability distribution calculation using complex
    annotated states, with imperfect purity and/or indistinguishability.

    Args:

        circuit (CompiledPhotonicCircuit) : The compiled circuit that is to be
                                    sampled from.

        inputs (dict) : The inputs to the system and their associated
                        probabilities.

        probability_func (Callable) : A method for calculation of a probability
            distribution, given a circuit and input state.

    Returns:

        dict : The calculated output probability distribution.

    """
    # Determine the input state combinations given the labels
    unique_inputs: set[State] = set()
    input_combinations: dict[AnnotatedState, list[State]] = {}
    for state in inputs:
        # Find all labels in a given state
        all_labels = []
        for mode in state:
            all_labels += mode
        # For all labels break them down into the corresponding states
        if all_labels:
            results = {lab: [0] * circuit.n_modes for lab in all_labels}
            for i, mode in enumerate(state):
                for m in mode:
                    results[m][i] += 1
            states = [State(s) for s in results.values()]
            unique_inputs |= set(states)
        else:  # Special case for empty annotated state
            states = [State([0] * circuit.n_modes)]
            unique_inputs.add(State([0] * circuit.n_modes))
        input_combinations[state] = states
    # For each of the unique inputs then need to work out the probability
    # distribution
    unique_results: dict[State, Any] = {}
    for in_state in unique_inputs:
        # Calculate sub distribution and store
        unique_results[in_state[: circuit.n_modes]] = probability_func(
            circuit, in_state
        )

    # Pre-calculate dictionary items to improve speed
    for r, pdist in unique_results.items():
        unique_results[r] = list(pdist.items())
    # Then combine the results above to work out the true output
    # probability for the inputs.
    stats_dict = {}
    for istate, combination in input_combinations.items():
        pdist = {}
        # Loop over states included in each combination
        for d in combination:
            if not pdist:
                pdist = dict(unique_results[d])
            else:
                new_pdist = {}
                # Combine existing distribution with new one
                for output, p1 in pdist.items():
                    for result, p2 in unique_results[d]:
                        new_state = output.merge(result)
                        if new_state not in new_pdist:
                            new_pdist[new_state] = p1 * p2
                        else:
                            new_pdist[new_state] += p1 * p2
                pdist = new_pdist
        # Then combine outputs and weight by input probability
        ip = inputs[istate]
        for ostate, op in pdist.items():
            if ostate not in stats_dict:
                stats_dict[ostate] = ip * op
            else:
                stats_dict[ostate] += ip * op
    return stats_dict



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import PhotonicCircuit, Unitary, random_unitary
from lightworks.interferometers import ErrorModel, Reck
from lightworks.interferometers.dists import Gaussian, TopHat


class TestReck:
    """
    Tests to check functionality of the Reck interferometer.
    """

    @pytest.mark.parametrize("n_modes", [2, 3, 7, 8, 15, 16])
    def test_equivalence(self, n_modes):
        """
        Checks map functionality produces an equivalent circuit for a range of
        mode values.
        """
        # Create test circuit
        test_circ = Unitary(random_unitary(n_modes))
        # Find mapped circuit
        mapped_circ = Reck().map(test_circ)
        # Then check equivalence
        assert (test_circ.U.round(8) == mapped_circ.U.round(8)).all()

    def test_sequential_maps(self):
        """
        Checks random procedure produces different circuits on subsequent calls
        when the random seed is not set.
        """
        # Create test circuit
        test_circ = Unitary(random_unitary(8))
        # Define error model with random variations
        emodel = ErrorModel()
        emodel.bs_reflectivity = Gaussian(0.5, 0.02, min_value=0, max_value=1)
        emodel.loss = TopHat(0.1, 0.2)
        emodel.phase_offset = Gaussian(0, 0.02)
        r = Reck(emodel)
        # Create two mapped circuits
        mapped_circ = r.map(test_circ)
        mapped_circ2 = r.map(test_circ)
        # Then check equivalence
        assert (mapped_circ.U.round(8) != mapped_circ2.U.round(8)).any()

    def test_random_seeding(self):
        """
        Checks random seeding produces repeatable circuits for a range of mode
        values.
        """
        # Create test circuit
        test_circ = Unitary(random_unitary(8))
        # Define error model with random variations
        emodel = ErrorModel()
        emodel.bs_reflectivity = Gaussian(0.5, 0.02, min_value=0, max_value=1)
        emodel.loss = TopHat(0.1, 0.2)
        emodel.phase_offset = Gaussian(0, 0.02)
        r = Reck(emodel)
        # Set seed and create two mapped circuits
        mapped_circ = r.map(test_circ, seed=12)
        mapped_circ2 = r.map(test_circ, seed=12)
        # Then check equivalence
        assert (mapped_circ.U.round(8) == mapped_circ2.U.round(8)).all()

    @pytest.mark.parametrize(
        "value", ["not_error_model", PhotonicCircuit(4), 0]
    )
    def test_error_model_invalid_type(self, value):
        """
        Checks that an exception is raised if the error_model is set to
        something other than an ErrorModel or None.
        """
        with pytest.raises(TypeError):
            Reck(error_model=value)

    def test_heralds(self):
        """
        Checks map functionality correctly moves heralds to new circuit.
        """
        # Create test circuit
        test_circ = Unitary(random_unitary(8))
        test_circ.herald((5, 5), 0)
        test_circ.herald((0, 3), (2, 1))
        # Check circuit can be mapped
        mapped_circ = Reck().map(test_circ)
        assert test_circ.heralds == mapped_circ.heralds



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import random

import numpy as np
import pytest

from lightworks.interferometers.dists import (
    Constant,
    Distribution,
    Gaussian,
    TopHat,
)
from lightworks.interferometers.dists.utils import is_number


class TestDistributions:
    """General tests for probability distribution (dists) module."""

    @pytest.mark.parametrize(
        "dist",
        [Constant(0.5), Gaussian(0.5, 0), TopHat(0.5, 0.5)],
    )
    def test_is_distribution(self, dist):
        """
        Checks each of the distribution classes is an instance of the
        distribution base class.
        """
        assert isinstance(dist, Distribution)

    @pytest.mark.parametrize(
        "value", [1, 0.5, np.inf, np.float64(1.2), [1, 0.5, np.inf]]
    )
    def test_is_number(self, value):
        """
        Confirms that is_number does not return an exception for valid values.
        """
        is_number(value)

    @pytest.mark.parametrize("value", ["1", None, True, [1, True]])
    def test_is_number_invalid(self, value):
        """
        Confirms that is_number returns an exception when an invalid value is
        attempted to be set.
        """
        with pytest.raises(TypeError):
            is_number(value)

    def test_custom_distribution_requires_value(self):
        """
        Checks that any created custom class which uses Distribution as the
        parent requires the value method for the class to be initialized.
        """

        class TestClass(Distribution):
            pass

        with pytest.raises(TypeError):
            TestClass()


class TestConstant:
    """Tests for Constant distribution"""

    def test_constant(self):
        """
        Checks that Constant distribution just returns set value.
        """
        val = random()
        c = Constant(val)
        # Check 100 times to ensure it always works
        for _i in range(100):
            assert val == c.value()

    def test_params_in_string_and_repr(self):
        """
        Checks that the assigned value is detailed in the string and repr.
        """
        val = random()
        c = Constant(val)
        assert str(val) in str(c)
        assert str(val) in repr(c)


class TestGaussian:
    """Tests for Constant distribution"""

    @pytest.mark.flaky(max_runs=3)
    def test_gaussian(self):
        """
        Checks that Gaussian distribution generates values with a mean close to
        the set value for large numbers.
        """
        dist = Gaussian(1, 0.2)
        vals = [dist.value() for _i in range(100000)]
        # Check within 5% of expected mean
        assert np.mean(vals) == pytest.approx(1, 0.05)

    def test_max_less_than_min(self):
        """
        Confirms that exception is raised if the max value is set lower than
        the min value
        """
        with pytest.raises(ValueError):
            Gaussian(1, 0.2, min_value=1, max_value=0)

    def test_values_within_set_bounds(self):
        """
        Checks that values are forced to exist within the set bounds by setting
        bounds to exist within a narrow window of the Gaussian distribution.
        """
        dist = Gaussian(1, 0.3, min_value=1, max_value=1.5)
        vals = [dist.value() for _i in range(100000)]
        # Check values are within the minimum and maximum range
        assert min(vals) >= 1
        assert max(vals) <= 1.5

    def test_random_seed(self):
        """
        Checks that random seed produce repeatable values.
        """
        dist = Gaussian(1, 0.3)
        # Set seed and sampler twice
        dist.set_random_seed(99)
        v1 = dist.value()
        dist.set_random_seed(99)
        v2 = dist.value()
        # Check equivalence
        assert v1 == v2

    def test_params_in_string_and_repr(self):
        """
        Checks that the assigned value is detailed in the string and repr.
        """
        val = random()
        dist = Gaussian(val, 2 * val)
        assert str(val) in str(dist)
        assert str(2 * val) in str(dist)
        assert str(val) in repr(dist)
        assert str(2 * val) in repr(dist)

    def test_params_in_string_and_repr_bounds(self):
        """
        Checks that the assigned min and max value is detailed in the string and
        repr.
        """
        val = random()
        dist = Gaussian(1, 0.5, min_value=val, max_value=2 * val)
        assert str(val) in str(dist)
        assert str(2 * val) in str(dist)
        assert str(val) in repr(dist)
        assert str(2 * val) in repr(dist)


class TestTopHat:
    """Tests for TopHat distribution"""

    def test_top_hat(self):
        """
        Checks that Top Hat distribution only creates values within the expected
        range.
        """
        dist = TopHat(0.4, 0.6)
        vals = [dist.value() for _i in range(100000)]
        # Check min and max value within set range
        assert min(vals) >= 0.4
        assert max(vals) <= 0.6

    def test_max_less_than_min(self):
        """
        Confirms that exception is raised if the max value is set lower than
        the min value
        """
        with pytest.raises(ValueError):
            TopHat(0.6, 0.4)

    def test_random_seed(self):
        """
        Checks that random seed produce repeatable values.
        """
        dist = TopHat(0.4, 0.6)
        # Set seed and sampler twice
        dist.set_random_seed(99)
        v1 = dist.value()
        dist.set_random_seed(99)
        v2 = dist.value()
        # Check equivalence
        assert v1 == v2

    def test_params_in_string_and_repr(self):
        """
        Checks that the assigned value is detailed in the string and repr.
        """
        val = random()
        dist = TopHat(val, 2 * val)
        assert str(val) in str(dist)
        assert str(2 * val) in str(dist)
        assert str(val) in repr(dist)
        assert str(2 * val) in repr(dist)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numpy import random

from lightworks.sdk.utils.random import process_random_seed

from .dists import Constant, Distribution


class ErrorModel:
    """
    Allows for configuration of an error model which can be applied to different
    components of the interferometers.
    """

    def __init__(self) -> None:
        self.bs_reflectivity = Constant(0.5)
        self.loss = Constant(0)
        self.phase_offset = Constant(0)

    def __str__(self) -> str:
        return (
            f"ErrorModel(bs_reflectivity = {self.bs_reflectivity}, "
            f"loss = {self.loss})"
        )

    def __repr__(self) -> str:
        return "lightworks.interferometers." + str(self)

    @property
    def bs_reflectivity(self) -> Distribution:
        """Returns currently in use beam splitter value distribution."""
        return self._bs_reflectivity

    @bs_reflectivity.setter
    def bs_reflectivity(self, distribution: Distribution) -> None:
        if not isinstance(distribution, Distribution):
            raise TypeError("bs_reflectivity should be a distribution object.")
        self._bs_reflectivity = distribution

    @property
    def loss(self) -> Distribution:
        """Returns currently in use loss value distribution."""
        return self._loss

    @loss.setter
    def loss(self, distribution: Distribution) -> None:
        if not isinstance(distribution, Distribution):
            raise TypeError("loss should be a distribution object.")
        self._loss = distribution

    @property
    def phase_offset(self) -> Distribution:
        """Returns currently in use phase_offset value distribution."""
        return self._phase_offset

    @phase_offset.setter
    def phase_offset(self, distribution: Distribution) -> None:
        if not isinstance(distribution, Distribution):
            raise TypeError("phase_offset should be a distribution object.")
        self._phase_offset = distribution

    def get_bs_reflectivity(self) -> float:
        """
        Returns a value for beam splitter reflectivity, which depends on the
        configuration of the error model.
        """
        return self._bs_reflectivity.value()

    def get_loss(self) -> float:
        """
        Returns a value for loss, which depends on the configuration of the
        error model.
        """
        return self._loss.value()

    def get_phase_offset(self) -> float:
        """
        Returns a value for phase offset, which depends on the configuration of
        the error model.
        """
        return self._phase_offset.value()

    def _set_random_seed(self, r_seed: int | None) -> None:
        """
        Set the random seed for the error_model to produce repeatable results.
        """
        seed = process_random_seed(r_seed)
        # Create a rng to modify the seed by, ensuring two distributions produce
        # different values
        rng = random.default_rng(seed)
        # Set random seed in each property if present
        for prop in [self._bs_reflectivity, self._loss, self._phase_offset]:
            if seed is not None:
                seed = int(rng.integers(2**31 - 1))
            prop.set_random_seed(seed)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.sdk.state import State
from lightworks.sdk.utils.exceptions import PhotonNumberError


def check_photon_numbers(states: list[State], target_n: int) -> None:
    """
    Raises an exception if photon numbers are mixed when running a
    simulation.
    """
    ns = [s.n_photons for s in states]
    if min(ns) != target_n or max(ns) != target_n:
        raise PhotonNumberError(
            "Mismatch in photon numbers between some inputs/outputs, "
            "this is not currently supported in the Simulator."
        )



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any

from lightworks.sdk.results import Result
from lightworks.sdk.state import State
from lightworks.sdk.tasks import Task, TaskData

from .caching import CacheData, check_parameter_updates, get_calculation_values

# ruff: noqa: D102


class EmulatorBackend(ABC):
    """
    Base class for all emulator backends. An outline of all possible functions
    should be included here.
    """

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def compatible_tasks(self) -> tuple[str, ...]: ...

    @abstractmethod
    def run(self, task: Task) -> Result[State, Any]: ...

    def _check_cache(self, data: TaskData) -> dict[str, Any] | None:
        name = data.__class__.__name__
        if hasattr(self, "_cache") and name in self._cache:
            old_values = self._cache[name].values
            new_values = get_calculation_values(data)
            if not check_parameter_updates(old_values, new_values):
                return self._cache[name].results
        # Return false if cache doesn't exist or name not found
        return None

    def _add_to_cache(self, data: TaskData, results: dict[str, Any]) -> None:
        if not hasattr(self, "_cache"):
            self._cache = {}
        name = data.__class__.__name__
        values = get_calculation_values(data)
        self._cache[name] = CacheData(values=values, results=results)



# Copyright 2024 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from types import FunctionType, NoneType

from lightworks.sdk.state import State


class PostSelectionType(ABC):
    """
    Base class for post-selection object.
    """

    @abstractmethod
    def validate(self, state: State | list[int]) -> bool:
        """Enforces all post-selection classes have the validate method."""


class PostSelection(PostSelectionType):
    """
    Post-selection

    Args:

        multi_rules (bool, optional) : Controls whether multiple rules can be
            applied to a single mode.

    """

    def __init__(self, multi_rules: bool = False) -> None:
        self.multi_rules = multi_rules
        self.__rules: list[Rule] = []
        self.__modes_with_rules: set[int] = set()

    @property
    def rules(self) -> list["Rule"]:
        """
        Returns currently applied post-selection rules.
        """
        return self.__rules

    @property
    def modes(self) -> list[int]:
        """
        Returns modes on which rules are currently applied.
        """
        return sorted(self.__modes_with_rules)

    def add(
        self, modes: int | tuple[int, ...], n_photons: int | tuple[int, ...]
    ) -> None:
        """
        Adds a post-selection rule across the provided modes, conditioning that
        the total number of photons across the modes is equal to any of the
        provided number of photons.

        Args:

            modes (int, tuple) : The modes across which post-selection should be
                applied. This can be a singular mode provided as an integer or
                a range of modes as a tuple.

            n_photons (int, tuple) : The valid photon number totals across the
                modes. This can be a singular integer value or a range of
                allowable photon numbers.

        """
        modes = check_int_or_tuple(modes)
        n_photons = check_int_or_tuple(n_photons)
        for values, msg in zip(
            (modes, n_photons), ("Mode numbers", "Photon numbers"), strict=True
        ):
            for v in values:
                if v < 0:
                    raise ValueError(msg + " cannot be negative.")
        # Check mode doesn't already have a rule if required
        if not self.multi_rules:
            for m in modes:
                if m in self.__modes_with_rules:
                    msg = (
                        f"Provided mode ({m}) already has a post-selection "
                        "rule applied. To allow this, set multi_rules attribute"
                        " to True."
                    )
                    raise ValueError(msg)
        self.__rules.append(Rule(modes, n_photons))
        for m in modes:
            self.__modes_with_rules.add(m)

    def validate(self, state: State | list[int]) -> bool:
        """
        Validates whether a provided State meets the set post-selection
        criteria.

        Args:

            state (State) : The state which is to be checked.

        Returns:

            bool : Indicates whether the provided State meets the post-selection
                criteria.

        """
        return all(rule.validate(state) for rule in self.__rules)


@dataclass(slots=True)
class Rule:
    """
    Stores a post-selection rule for a number of modes and n_photons.
    """

    modes: tuple[int, ...]
    n_photons: tuple[int, ...]

    def as_tuple(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
        """Returns a tuple representation of the post-selection rule."""
        return (self.modes, self.n_photons)

    def validate(self, state: State | list[int]) -> bool:
        """Validates a provided state meets the post-selection rule."""
        return sum(state[m] for m in self.modes) in self.n_photons


class PostSelectionFunction(PostSelectionType):
    """
    Allows for post-selection to be implemented with a provided function.
    """

    def __init__(self, function: Callable[[State], bool]) -> None:
        if not isinstance(function, FunctionType):
            raise TypeError("Post-selection not a function.")
        self.__function = function

    def validate(self, state: State | list[int]) -> bool:
        """
        Validates whether a provided State meets the set post-selection
        criteria.

        Args:

            state (State) : The state which is to be checked.

        Returns:

            bool : Indicates whether the provided State meets the post-selection
                criteria.

        """
        return self.__function(state)


class DefaultPostSelection(PostSelectionType):
    """
    Provides a default post-selection function which always returns True.
    """

    def validate(self, state: State | list[int]) -> bool:  # noqa: ARG002
        """
        Will return True regardless of provided state.
        """
        return True


def check_int_or_tuple(value: int | Sequence[int]) -> tuple[int, ...]:
    """
    Process a provided value, if it is a sequence then it will check all values
    are integers, otherwise it will validate the value and convert to a tuple.
    """
    if isinstance(value, Sequence):
        value = list(value)
        for i, v in enumerate(value):
            value[i] = check_int(v)
        return tuple(value)
    value = check_int(value)
    return (value,)


def check_int(value: int) -> int:
    """
    Processes a provided single value, trying to convert to an integer if not
    already and and then returning this.
    """
    if not isinstance(value, int):
        if value == int(value):
            value = int(value)
        else:
            raise ValueError(
                "Provided value should be a tuple of integers or an integer."
            )
    return value


def process_post_selection(
    post_selection: PostSelectionType | Callable[[State], bool] | None,
) -> PostSelectionType | None:
    """
    Takes a provided post-selection value and converts this into one of the
    PostSelection objects with a validate method.

    Args:

        post_selection (PostSelectionType | Callable | None) : The
            post-selection value to be processed. If a callable this will be
            converted into the PostSelectionFunction object, otherwise if None
            then the DefaultPostSelection object will be used.

    Returns:

        PostSelectionType | None : The created PostSelection object or None, if
            the original value wasn't one of these types.

    """
    if isinstance(post_selection, FunctionType):
        post_selection = PostSelectionFunction(post_selection)
    elif not isinstance(post_selection, PostSelectionType | NoneType):
        raise TypeError(
            "Provided post_select value should be a PostSelection object, "
            "function or None."
        )
    return post_selection
